#! /usr/bin/env python3
from dolfin import *
from anba4 import *
import argparse
import multiprocessing
import json
import numpy as np
import yaml
import os
import pyvista as pv
from functools import partial


def get_material_db(material_map):
    assert os.path.isfile(material_map)
    mm = json.load(open(material_map, "r"))
    gdir = os.path.dirname(material_map)

    mat_db = None
    if "matdb" in mm:  # check if the material map file points to a material db
        mat_db = yaml.safe_load(open(os.path.join(gdir, mm["matdb"])))
        # , Loader=yaml.CLoader)

        # check if there is a -1 material in the matdb, and assign it
        # this material ID is used by the section mesher for the bondlines
        # that connect the webs to the shell
        if "-1" in mat_db:
            mm["-1"] = -1
    else:
        exit(
            "material map available, but no link to material db, need matdb definition to do FEA"
        )

    mm_inv = {v: k for k, v in mm.items()}

    materials = {}
    for i in mm_inv:
        if i != mm["matdb"]:
            matdb_entry = mat_db[mm_inv[i]]
            if "tEx" in matdb_entry:  # ortho material
                matMechanicProp = np.zeros((3, 3))
                matMechanicProp[0, 0] = matdb_entry["tEz"] * 1e6  # e_xx
                matMechanicProp[0, 1] = matdb_entry["tEy"] * 1e6  # e_yy
                matMechanicProp[0, 2] = matdb_entry["tEx"] * 1e6  # e_zz

                matMechanicProp[1, 0] = matdb_entry["tGxz"] * 1e6  # g_yz
                matMechanicProp[1, 1] = matdb_entry["tGxy"] * 1e6  # g_xz
                matMechanicProp[1, 2] = matdb_entry["tGyz"] * 1e6  # g_xy

                matMechanicProp[2, 0] = matdb_entry["tnuxz"]  # nu_zy
                matMechanicProp[2, 1] = matdb_entry["tnuxy"]  # nu_zx
                matMechanicProp[2, 2] = matdb_entry["tnuyz"]  # nu_xy

                materials[i] = material.OrthotropicMaterial(
                    matMechanicProp, matdb_entry["rho"]
                )
            else:
                materials[i] = material.IsotropicMaterial(
                    [
                        (
                            matdb_entry["E"] * 1e6
                            if "E" in matdb_entry
                            else matdb_entry["Ex"] * 1e6
                        ),
                        matdb_entry["nu"],
                    ],
                    matdb_entry["rho"],
                )

    return materials


def run_mesh(meshname, matdb):
    print(f"run {meshname}")

    infile = XDMFFile(meshname)
    mesh = Mesh()
    infile.read(mesh)

    pvmesh = pv.read(meshname)

    # matid = MeshFunction("size_t", mesh, mesh.topology().dim())

    # Basic material parameters. 9 is needed for orthotropic materials.
    # TODO materials and orientations
    # Meshing domain.
    materials = MeshFunction("size_t", mesh, mesh.topology().dim())
    fiber_orientations = MeshFunction("double", mesh, mesh.topology().dim())
    plane_orientations = MeshFunction("double", mesh, mesh.topology().dim())

    matuniq = np.unique(pvmesh.cell_data["mat"])  # material indices (from material_map)

    # map for materials in anba (index in matuniq)
    mat_map_0 = dict(zip(matuniq, range(len(matuniq))))

    matids = [mat_map_0[i] for i in pvmesh.cell_data["mat"].tolist()]

    plane_angles = list(pvmesh.cell_data["angle2"])

    materials.set_values(matids)

    # TODO, doesn't work for off axis laminates for now
    fiber_orientations.set_all(0.0)

    # transverse orientations
    plane_orientations.set_values(plane_angles)

    # Build material property library.

    matLibrary = [matdb[i] for i in matuniq]

    anba = anbax(mesh, 2, matLibrary, materials, plane_orientations, fiber_orientations)
    stiff = anba.compute()

    stiffness_matrix = stiff.getDenseArray()

    mass = anba.inertia()

    mass_matrix = mass.getDenseArray()

    decoupled_stiff = DecoupleStiffness(stiff)

    angle = PrincipalAxesRotationAngle(decoupled_stiff)

    mass_center = ComputeMassCenter(mass)
    tension_center = ComputeTensionCenter(stiffness_matrix)
    shear_center = ComputeShearCenter(stiffness_matrix)
    output = {
        "name": meshname,
        "stiffness": stiffness_matrix.tolist(),
        "mass_matrix": mass_matrix.tolist(),
        "decoupled_stiffness": decoupled_stiff.tolist(),
        "principal_axes_rotation": angle,
        "mass_center": mass_center,
        "tension_center": tension_center,
        "shear_center": shear_center,
    }

    with open(f"{meshname}.json", "w") as write_file:
        json.dump(output, write_file, indent=4)


def main():
    p = argparse.ArgumentParser(description="run a series of sections through anba4")
    p.add_argument("meshes", nargs="*")
    p.add_argument("matdb", help="material map json file")
    p.add_argument("--debug", action="store_true")
    args = p.parse_args()

    mdb = get_material_db(args.matdb)

    part = partial(run_mesh, matdb=mdb)
    if args.debug:
        for i in args.meshes:
            part(i)
    else:
        p = multiprocessing.Pool()
        # run async seems to avoid a sporadic blocking error
        # p.map(part, args.meshes)
        r = p.map_async(part, args.meshes)
        r.wait()


if __name__ == "__main__":
    main()



#! /usr/bin/env python

# from ruamel import YAML as YAML
from ruamel.yaml import YAML
import argparse
from matplotlib import pyplot as plt
import numpy as np


def plot_planform(bl, prefix):
    odct = {}
    absolute_thickness = np.array(bl["chord"]["values"]) * np.array(
        bl["rthick"]["values"]
    )
    fig, ax = plt.subplots(5, 2, figsize=(10, 12))
    c = 0
    bl["twist"]["values"] = np.degrees(bl["twist"]["values"]).tolist()
    for i in bl:
        if "grid" in bl[i] and "values" in bl[i]:
            ax[c][0].plot(bl[i]["grid"], bl[i]["values"])
            ax[c][0].set_title(i)
            odct[i] = [list(j) for j in zip(bl[i]["grid"], bl[i]["values"])]
            c += 1

    ax[c][0].plot(bl["chord"]["grid"], absolute_thickness)

    for c, i in enumerate(bl["reference_axis"]):
        ax[c][1].plot(
            bl["reference_axis"][i]["grid"], bl["reference_axis"][i]["values"]
        )
        ax[c][1].set_title(i)
        odct[("d" if i in ["x", "y"] else "") + i] = [
            list(j)
            for j in zip(
                bl["reference_axis"][i]["grid"], bl["reference_axis"][i]["values"]
            )
        ]
    of = f"{prefix}_planform"
    fig.savefig(of)
    print(f"written plot to {of}")

    odct["thickness"] = odct["rthick"]
    del odct["rthick"]
    return odct


def export_airfoils(af, prefix):
    x, ax = plt.subplots(1, 1, figsize=(12, 12))

    af_dict = {}
    for i in af:
        fn = f'airfoils/af_{i["name"]}.dat'
        buf = "%s\n" % i["name"]
        for j in zip(i["coordinates"]["x"], i["coordinates"]["y"]):
            buf += "%s %s\n" % j
        open(fn, "w").write(buf)

        ax.plot(i["coordinates"]["x"], i["coordinates"]["y"], label=fn)

        af_dict[i["relative_thickness"]] = fn
    ax.legend(loc="best")
    plt.savefig(f"{prefix}_airfoils.png")

    return af_dict


def add_rthick_if_absent(x):
    shp = x["components"]["blade"]["outer_shape_bem"]

    if "rthick" not in shp:
        af = x["airfoils"]

        thick = dict([(i["name"], i["relative_thickness"]) for i in af])

        rthick = [thick[i] for i in shp["airfoil_position"]["labels"]]

        shp["rthick"] = {
            "grid": shp["chord"]["grid"],
            "values": np.interp(
                shp["chord"]["grid"], shp["airfoil_position"]["grid"], rthick
            ).tolist(),
        }

    return x


def plot_laminates(model, prefix):
    lam = model["components"]["blade"]["internal_structure_2d_fem"]

    fig, ax = plt.subplots(1, 1, figsize=(12, 12))

    for i in lam["layers"]:
        ax.plot(i["thickness"]["grid"], i["thickness"]["values"], label=i["name"])

    ax.legend(loc="best")

    fig.savefig(f"{prefix}_lams")


if __name__ == "__main__":
    p = argparse.ArgumentParser(
        description="read a windio file and convert to b3p input, plot airfoils, planform and composites"
    )
    p.add_argument("yaml", help="WindIO yaml input file")
    p.add_argument("template", help="b3p template yaml file")
    p.add_argument("-o", default="out")
    args = p.parse_args()

    yaml = YAML(typ="rt")
    template = yaml.load(open(args.template, "r"))
    # , Loader=yaml.RoundTripLoader)

    x = yaml.load(open(args.yaml, "r"))

    x = add_rthick_if_absent(x)

    out_var = plot_planform(x["components"]["blade"]["outer_shape_bem"], args.o)

    af_dict = export_airfoils(x["airfoils"], args.o)

    for i in out_var:
        template["planform"][i] = out_var[i]
        # print(i, out_var[i])

    template["aero"]["airfoils"] = af_dict

    #
    plot_laminates(x, args.o)

    yaml.dump(
        dict(template),
        open(f"{args.o}.yml", "w"),
        # Dumper=yaml.RoundTripDumper,
        # default_flow_style=None,
    )



#! /usr/bin/env python

from utils import yml_portable
import fire
import numpy as np
import os
import pandas as pd
from matplotlib import pyplot as plt
from scipy.optimize import fmin
import numpy as np
from ccblade.ccblade import CCAirfoil, CCBlade
import math


def load_polar(pname):
    "load a polar with a given name, interpolate to set alpha range"
    print("loading polar", pname)
    if not os.path.isfile(pname):
        raise IOError(f"Polar {pname} not found")

    with open(pname, "r") as f:
        lines = f.readlines()

    read_start_index = next(
        i for i, line in enumerate(lines) if line.startswith("-180")
    )
    read_end_index = next((i for i, line in enumerate(lines) if "EOT" in line), None)

    lst = [
        [float(j) for j in line.split()]
        for line in lines[read_start_index:read_end_index]
    ]
    g = list(zip(*lst))
    alpha_new = (
        list(np.linspace(-180, -20, 25))
        + list(np.linspace(-19, 20, 29))
        + list(np.linspace(21, 180, 25))
    )
    cl = np.interp(alpha_new, g[0], g[1])
    cd = np.interp(alpha_new, g[0], g[2])
    cm = np.interp(alpha_new, g[0], g[3])
    return [alpha_new, cl, cd, cm]


def plot_interpolated_polars(t, data, of="polars.png"):
    fig, ax = plt.subplots(3, 1, figsize=(12, 19))
    for n, i in enumerate(t):
        alpha, cl, cd, cm = data[:, :, n]
        ax[0].plot(alpha, cl, label=f"t={t[n]:.3f}")
        ax[1].plot(alpha, cd)
        ax[2].plot(cd, cl)
        ax[0].set_ylabel("Cl")
        ax[1].set_ylabel("Cd")
        ax[2].set_ylabel("Cm")
        ax[0].legend()
    fig.savefig(of)


def plot_polars(polars, of="polars_in.png"):
    fig, ax = plt.subplots(3, 1, figsize=(6, 8))
    for i in polars:
        alpha, cl, cd, cm = i[1]  # polars[i]
        ax[0].plot(alpha, cl, label=i[0])
        ax[1].plot(alpha, cd)
        ax[2].plot(cd, cl)
        ax[0].set_ylabel("Cl")
        ax[1].set_ylabel("Cd")
        ax[2].set_ylabel("Cl - Cd")
        ax[0].legend()
    fig.savefig(of)


def interpolate_polars(polars, tnew, of=None):
    indices = range(len(polars[0][1][0]))
    t = [polar[0] for polar in polars]
    data = np.array(
        [
            [
                np.interp(
                    np.flip(tnew.values),
                    np.flip(t),
                    np.flip([polar[1][i][idx] for polar in polars]),
                )
                for idx in indices
            ]
            for i in range(4)
        ]
    )
    output_polars = [
        CCAirfoil(
            Re=[1e6],
            alpha=data[0, :, i],
            cl=data[1, :, i],
            cd=data[2, :, i],
            cm=data[3, :, i],
        )
        for i in reversed(range(len(tnew)))
    ]
    if of:
        plot_polars(polars, of.replace(".png", "_in.png"))
        plot_interpolated_polars(tnew, data, of)

    return output_polars


class RotorOptimizer:
    def __init__(self, rotor, uinf, rated_power=20e7, omega=None, maxiter=5):
        self.rotor = rotor
        self.uinf = uinf
        self.omega = omega
        self.rated_power = rated_power
        self.cache = {}
        self.latest = None
        self.maxiter = maxiter

    def evaluate(self, x, coefficients=False):
        if self.omega == None:
            omega, pitch = x
        else:
            omega, pitch = self.omega, x[0]

        if (omega, pitch) not in self.cache:
            outputs, _ = self.rotor.evaluate(
                self.uinf, omega, pitch, coefficients=coefficients
            )
            self.cache[(omega, pitch)] = outputs

        P = np.mean(self.cache[(omega, pitch)]["P"])
        return P, self.cache[(omega, pitch)]

    def objective(self, x):
        ev, _ = self.evaluate(x)
        return np.fabs(self.rated_power - ev)  # + np.sum(penalties)

    def optimize(self, initial_guess):
        result = fmin(self.objective, initial_guess, maxiter=self.maxiter)

        rr, rdet = self.evaluate(result)
        # print(f"{self.uinf} {self.omega} {self.rated_power} result {result}")
        # print(f"result {rr} {rdet}")
        return result


def tsr2omega(tsr, uinf, radius, max_tip_speed=95.0):
    ts = np.minimum(uinf * tsr, max_tip_speed)
    return (ts * 60.0) / (2.0 * np.pi * radius)


def omega2tsr(omega, uinf, radius):
    return omega * 2.0 * np.pi * radius / (uinf * 60.0)


def plot_grid(num_plots, figsize=(15, 15)):
    # Determine grid dimensions (try to get as square as possible)
    grid_size = math.isqrt(num_plots)

    # Adjust grid dimensions if necessary
    if grid_size * grid_size < num_plots:
        columns = grid_size
        rows = np.ceil(num_plots / grid_size).astype(int)
    else:
        rows = columns = grid_size

    fig, axs = plt.subplots(rows, columns, figsize=figsize)
    axs = axs.flatten()

    # Remove unused subplots
    for idx in range(len(axs) - 1, rows * columns):
        # print("deleting", idx)
        fig.delaxes(axs[idx])

    return fig, axs  # , rows, columns


def find_closest_x(x_values, evaluations, target, order):
    assert len(x_values) == len(evaluations)
    # fit a polynomial of given order
    poly_coeffs = np.polyfit(x_values, evaluations, order)
    # create a polynomial function from the coefficients
    poly = np.poly1d(poly_coeffs)
    # create a dense range of x values for interpolation
    x_dense = np.linspace(x_values[0], x_values[-1], 10000)
    # find the interpolated x value where the absolute difference to the target is smallest
    x_closest = x_dense[np.argmin(np.abs(poly(x_dense) - target))]
    return x_closest


def plot_bladeloads(r, data_dict, of="bladeloads.png"):
    fig, axs = plot_grid(len(data_dict), figsize=(15, 15))
    # Iterate over the dictionary and plot each array
    for idx, (name, array) in enumerate(data_dict.items()):
        axs[idx].plot(r, array)
        axs[idx].set_title(name)

    fig.tight_layout()
    fig.savefig(of)
    print(f"Saved {of}")


class controloptimize:
    def __init__(self, rotor, max_tipspeed, rtip, rating, uinf, workdir):
        self.rotor = rotor
        self.max_tipspeed = max_tipspeed
        self.rating = rating
        self.uinf = uinf
        self.rtip = rtip
        self.workdir = workdir

    def control_opt_below_rated(
        self, starting_uinf=6, starting_tsr=10, starting_pitch=0
    ):
        """
        Optimize the control below rated power, optimize for both tsr and pitch, do this only for one wind speed (default 6m/s)

        starting_uinf: wind speed to optimize for
        starting_tsr: initial guess for tsr
        starting_pitch: initial guess for pitch
        """
        omega = tsr2omega(
            starting_tsr, uinf=6, radius=self.rtip, max_tip_speed=self.max_tipspeed
        )
        optimizer = RotorOptimizer(self.rotor, [starting_uinf])
        initial_guess = np.array([omega, starting_pitch])
        optimal_values = optimizer.optimize(initial_guess)
        print(f"initial guess {initial_guess} optimal values {optimal_values}")
        init_val = optimizer.evaluate(initial_guess)[0]
        opt_val, optt = optimizer.evaluate(optimal_values, coefficients=True)

        print(f" {init_val} {opt_val}, improvement {opt_val/init_val}")
        self.optimal_tsr = omega2tsr(optimal_values[0], starting_uinf, self.rtip)
        self.fine_pitch = optimal_values[1]

        loads, _ = self.rotor.distributedAeroLoads(
            self.uinf, optimal_values[0], optimal_values[1], 0
        )
        plot_bladeloads(
            self.rotor.r, loads, of=os.path.join(self.workdir, "bladeloads.png")
        )
        print(f"optimal tsr {self.optimal_tsr} {self.fine_pitch}")

    def control_opt_above_rated(self):
        # first run through the powercurve with fine_pitch all the way
        self.omega = tsr2omega(
            self.optimal_tsr,
            uinf=self.uinf,
            radius=self.rtip,
            max_tip_speed=self.max_tipspeed,
        )
        self.tsr = omega2tsr(self.omega, uinf=self.uinf, radius=self.rtip)
        self.pitch = self.fine_pitch * np.ones_like(self.uinf)

        init_pc, _ = self.rotor.evaluate(
            self.uinf,
            self.omega,
            self.pitch,
            coefficients=False,
        )
        rotorplot(init_pc, self.uinf, of=os.path.join(self.workdir, "init.png"))

        # find the wind speeds that are over rated power
        overrated = np.where(init_pc["P"] > self.rating)
        print(f"overrated {overrated}, {self.uinf[overrated]}")
        upost = self.uinf[overrated]
        ompost = self.omega[overrated]
        pitch_over_rated = []
        closest_pitch = self.fine_pitch

        # keep tsr fixed and find the pitch that gives the rated power
        for i in np.array(list(zip(upost, ompost))):
            pitch = np.linspace(closest_pitch, closest_pitch + 15.0, 8)
            ui = np.ones_like(pitch) * i[0]
            oi = np.ones_like(pitch) * i[1]
            pg, _ = self.rotor.evaluate(ui, oi, pitch, coefficients=False)
            closest_pitch = find_closest_x(pitch, pg["P"], self.rating, 3)
            pitch_over_rated.append(closest_pitch)

        self.pitch[overrated] = pitch_over_rated

        out_pc, _ = self.rotor.evaluate(
            self.uinf,
            self.omega,
            self.pitch,
            coefficients=True,
        )
        out_pc["omega"] = self.omega
        out_pc["tsr"] = self.tsr
        out_pc["pitch"] = self.pitch

        # plot the result
        rotorplot(
            out_pc,
            self.uinf,
            labels=["P", "CP", "Mb", "T", "omega", "pitch", "tsr"],
            of=os.path.join(self.workdir, "out.png"),
        )
        print(out_pc.keys())
        print(f"pitch {self.pitch}")
        return out_pc


def rotorplot(op, uinf, labels=["P", "CP", "T", "Mb"], of="__temp.png"):
    lab = [i for i in labels if i in op]

    # fig, ax = plt.subplots(len(lab), 1, figsize=(10, 10))
    fig, ax = plot_grid(len(lab), figsize=(10, 10))
    # print(len(ax), len(lab), rows, columns)
    for n, i in enumerate(lab):
        ax[n].plot(uinf, op[i], label=f"{i} max={op[i].max():.2f}")
        ax[n].legend()
        ax[n].grid()
        ax[n].set_ylabel(i)

    fig.tight_layout()
    fig.savefig(of)
    print(f"saved {of}")


class ccblade_run:
    def __init__(self, blade):
        self.dct = yml_portable.yaml_make_portable(blade)
        workdir = self.dct["general"]["workdir"]
        bem = {
            "rated_power": 10e6,
            "polars": {},
            "B": 3,
            "rho": 1.225,
            "tilt": 5,
            "precone": 3,
            "shearExp": 0.1,
            "hubHt": 140.0,
            "mu": 1.81206e-5,
            "yaw": 0,
            "max_tipspeed": 95.0,
            "uinf": [3, 5, 7, 9, 10, 11, 12, 13, 16, 20],
        }
        missing_keys = [key for key in bem if key not in self.dct["aero"]["bem"]]

        print(
            f"the following keys are not specified in aero/bem: {missing_keys}, using defaults {bem}"
        )
        bem |= self.dct["aero"]["bem"]

        self.prefix = os.path.join(
            self.dct["general"]["workdir"], self.dct["general"]["prefix"]
        )

        if "polars" not in bem:
            exit("no polars in blade file")

        if not os.path.isfile(f"{self.prefix}.pck"):
            exit("blade not run yet")

        plf = pd.read_csv(f"{self.prefix}_sca_50.csv", sep=";")

        plrs = sorted(
            [(i[0], load_polar(i[1])) for i in bem["polars"].items()],
            reverse=True,
        )
        iplr = interpolate_polars(
            plrs, plf.relative_thickness, of=os.path.join(workdir, "polars.png")
        )
        rhub, rtip = plf.z.iloc[0], plf.z.iloc[-1]
        self.rotor = CCBlade(
            plf.z - plf.z[0],
            plf.chord,
            plf.twist,
            iplr,
            rhub,
            rtip,
            B=bem["B"],
            rho=bem["rho"],
            mu=bem["mu"],
            precone=bem["precone"],
            tilt=bem["tilt"],
            yaw=bem["yaw"],
            shearExp=bem["shearExp"],
            hubHt=bem["hubHt"],
            derivatives=False,
        )

        print(f"Rotor from {rhub} to {rtip}")

        copt = controloptimize(
            self.rotor,
            bem["max_tipspeed"],
            rtip,
            bem["rated_power"],
            uinf=np.array(bem["uinf"]),
            workdir=workdir,
        )

        copt.control_opt_below_rated()

        copt.control_opt_above_rated()

    def __str__(self):
        return ""


# def main(blade):
#     ccblade_run(blade=blade)


# def ccblade_run_test():
#     ccblade_run()


# if __name__ == "__main__":
#     fire.Fire(main)


def main():
    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(ccblade_run)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import pandas as pd
import pyvista
import numpy as np
import fire
import glob
import os
import re
from b3p import frd2vtu


def has_later_vtu(frd):
    output = frd.replace(".frd", ".vtu")
    # print(output, os.path.exists(output))
    if not os.path.exists(output):
        return False
    frd_time = os.path.getmtime(frd)
    vtu_time = os.path.getmtime(output)
    return frd_time < vtu_time


def all_frd2vtu(workdir):
    for i in glob.glob(f"{workdir}/*.frd"):
        print(f"processing {i}")
        if not has_later_vtu(i):
            frd2vtu(i)
        else:
            print(f"skipping {i}")


def digitize_strain_distribution(z, strain, num_bins=100):
    # Check if strain has multiple components
    if strain.ndim == 1:
        strain = strain.reshape(-1, 1)
    elif strain.ndim != 2:
        raise ValueError("Invalid strain input. Expected 1D or 2D array.")

    num_components = strain.shape[1]

    # Bin the values in the first column into a specified number of bins
    zbin = np.linspace(z.min(), z.max(), num=num_bins + 1)
    bins = np.digitize(z, bins=zbin)

    # Find the maximum and minimum value in the second column for each bin and component
    max_vals = np.zeros((num_bins, num_components))
    min_vals = np.zeros((num_bins, num_components))

    for bin_num in range(1, num_bins + 1):
        for comp in range(num_components):
            max_vals[bin_num - 1, comp] = np.max(strain[bins == bin_num, comp])
            min_vals[bin_num - 1, comp] = np.min(strain[bins == bin_num, comp])

    return 0.5 * (zbin[:-1] + zbin[1:]), min_vals, max_vals


class ccx2vtu:
    def __init__(self, workdir, wildcard=""):
        self.workdir = workdir
        self.frds = glob.glob(f"{workdir}/*lc*{wildcard}*.frd")

    def load_grids(self):
        print(f"** loading grids {self.frds}")
        self.grids = {}
        for i in self.frds:
            self.grids[i] = (
                pyvista.UnstructuredGrid(i.replace(".frd", ".vtu"))
                if has_later_vtu(i)
                else frd2vtu.frd2vtu(i, multi=True)
            )

    def tabulate(self, n_bins=50):
        # Initialize an empty list to store the results
        results = []
        if len(self.grids) == 0:
            print(f"** no grids found in {self.workdir}")
            return None

        # Loop through the file_list
        for grid in self.grids:
            # Read the input data from the current file
            input_data = self.grids[grid]  # pd.read_csv(file_name)
            strain0 = next(
                (s for s in input_data.point_data.keys() if s.startswith("strain")),
                None,
            )

            z = input_data.points[:, 2]  # input_data['z'].values
            strain = input_data.point_data[strain0]

            # Compute the digitized strain distribution for the current input
            zbin, min_vals, max_vals = digitize_strain_distribution(
                z, strain, num_bins=n_bins
            )

            # Combine min_vals and max_vals into a single array
            vals = np.hstack([min_vals, max_vals])

            # Create MultiColumns for the current input
            # print(grid)
            columns = pd.MultiIndex.from_product(
                [
                    [re.search(r"_lc_(.+)\.frd", grid)[1]],
                    ["e_min", "e_max"],
                    ["xx", "yy", "zz", "xy", "yz", "xz"],
                ],
                names=["file", "type", "comp"],
            )

            # Create a DataFrame for the current input
            df = pd.DataFrame(vals, index=zbin, columns=columns)

            of = grid.replace(".frd", "_eps2d.pq")
            df.to_parquet(of)
            print(f"** written to {of}")

            # Append the current DataFrame to the results list
        #     results.append(df)

        # # Concatenate all the DataFrames in the results list
        # result_df = pd.concat(results)

        # return result_df


if __name__ == "__main__":
    fire.Fire(ccx2vtu)



from b3p import splining
from b3p import loft_utils
from b3p import blade_section

import pandas as pd
import numpy as np
from copy import deepcopy as dc
from matplotlib import pyplot as plt
import pickle
import math
import pyvista as pv


class blade:
    def __init__(
        self,
        chord,
        thickness,
        twist,
        dx,
        dy,
        z,
        airfoils,
        chordwise_sampling,
        np_spanwise=100,
    ):
        """Build a blade geometry from a yaml file
        :param chord: list of chord lengths
        :param thickness: list of relative thicknesses
        :param twist: list of twist angles
        :param dx: list of x offsets
        :param dy: list of y offsets
        :param z: list of z offsets
        :param airfoils: dict of airfoils
        :param chordwise_sampling: list of chordwise sampling points
        :param np_spanwise: number of spanwise points"""
        self.np_spanwise = np_spanwise
        self.np_chordwise = len(chordwise_sampling)
        self._load_airfoils(airfoils, chordwise_sampling)
        self._interpolate_planform(chord, thickness, twist, dx, dy, z)
        self._place_airfoils()

    def to_table(self, x, prefix="prebend_out"):
        """
        write out a table of the blade

        parameters:
            x (list) : list of sections where airfoils are to be interpolated
            prefix (str) : prefix for output file
        """
        cols = [
            "relative_r",
            "z",
            "prebend",
            "chord",
            "relative_thickness",
            "absolute_thickness",
            "twist",
        ]
        df = pd.DataFrame(
            np.array(
                [x]
                + [
                    np.interp(x, i[0], i[1])
                    for i in [
                        self.z,
                        self.dx,
                        self.chord,
                        self.thickness,
                        self.absolute_thickness,
                        self.twist,
                    ]
                ]
            ).T,
            columns=cols,
        )
        df.to_csv(f"{prefix}.csv", index=False, sep=";")

    def _load_airfoils(self, airfoils, x):
        """
        fill self.airfoil by interpolating from input airfoil set

        args:
            airfoils (dict) : airfoils dict

            x (list) : list of sections where airfoils are to be interpolated
        """
        print("** loading airfoils")
        self.airfoils = {}
        for i in sorted(airfoils):
            if type(airfoils[i]) == str:
                if airfoils[i].find("du") != -1:
                    print(f"** load {airfoils[i]} normalised")
                    t = loft_utils.load(airfoils[i], normalise=True)
                else:
                    print(f"** load {airfoils[i]} unnormalised")
                    t = loft_utils.load(airfoils[i], normalise=False)  # fix so we don't
                    # normalise flatback root
            else:
                # load an airfoil from the self-contained format, which is a dict with keys xy
                print(f"** loading airfoil {airfoils[i]['name']} at thickness {i}")
                t = airfoils[i]["xy"]

            self.airfoils[i] = loft_utils.interp(x, t)[:2]

    def _interpolate_planform(self, chord, thickness, twist, dx, dy, z):
        """
        spline the planform based on parameters

        parameters:
            chord (list) : chord distribution [(r,chord),..]
            thickness (list) : thickness distribution [(r,t),..] --> relative
            twist (list) : twist distribution [(r,twist),..]
            dx (list) : x distribution [(r,x),..]
            dy (list) : y distribution [(r,y),..]
            z (list) : z distribution [(r,z),..]
            flatten_lw (bool): flag indicating whether to run LW side flattening
        """
        self.x = np.linspace(0, 1.0, self.np_spanwise)
        (
            self.input_chord,
            self.input_thickness,
            self.input_twist,
            self.input_dx,
            self.input_dy,
        ) = (
            list(zip(*dc(chord))),
            list(zip(*dc(thickness))),
            list(zip(*dc(twist))),
            list(zip(*dc(dx))),
            list(zip(*dc(dy))),
        )

        self.chord = splining.intp_c(self.x, chord)
        self.twist = splining.intp_c(self.x, twist)
        self.thickness = splining.intp_c(self.x, thickness)
        self.dx = splining.intp_c(self.x, dx)
        self.dy = splining.intp_c(self.x, dy)
        self.z = splining.intp_c(self.x, z)
        self.absolute_thickness = [
            self.x,
            [i[0] * i[1] for i in zip(self.chord[1], self.thickness[1])],
        ]

    def plot(self, name="_dum", fname="_out.png"):
        """
        plot parameter distributions

        args:
            name (str) : blade candidate name

            fname (str): plot output file name

        """
        plt.figure(figsize=(15, 15))
        plt.subplot(3, 3, 1)
        plt.title("chord rotor_diam=%.3f" % (2.0 * max(self.z[1])))
        plt.plot(self.x, self.chord[1], label=name)
        plt.plot(self.input_chord[0], self.input_chord[1], "o", label=f"{name}_input")
        plt.xlabel("rel span (-)")
        plt.ylabel("chord (m)")
        plt.grid(True)
        plt.legend(loc="best").get_frame().set_alpha(0.5)

        plt.subplot(3, 3, 8)
        plt.plot(self.x, self.chord[1], label=name)
        plt.plot(self.input_chord[0], self.input_chord[1], "o", label=f"{name}_input")
        plt.grid(True)
        plt.xlim(0.9, 1)

        plt.title("tip chord")
        plt.subplot(3, 3, 2)
        plt.plot(self.x, self.twist[1], label=name)
        plt.plot(self.input_twist[0], self.input_twist[1], "o", label=f"{name}_input")
        self.title_plot("twist", 3)
        plt.plot(self.x, self.thickness[1], label=name)
        plt.plot(
            self.input_thickness[0],
            self.input_thickness[1],
            "o",
            label=f"{name}_input",
        )
        self.title_plot("rel thickness", 4)
        plt.plot(self.absolute_thickness[0], self.absolute_thickness[1], label=name)
        self.title_plot("abs thickness", 7)
        plt.plot(self.absolute_thickness[0], self.absolute_thickness[1], label=name)
        plt.title("abs thickness")
        plt.grid(True)
        plt.xlim(0, 0.2)

        plt.subplot(3, 3, 5)
        plt.plot(self.dx[0], self.dx[1], label=f"{name}_x")
        plt.plot(self.input_dx[0], self.input_dx[1], "o", label=f"{name}_input")
        plt.plot(self.dy[0], self.dy[1], label=f"{name}_y")
        plt.plot(self.input_dy[0], self.input_dy[1], "o", label=f"{name}_input")
        plt.legend(loc="best").get_frame().set_alpha(0.5)
        plt.savefig(fname, dpi=100)

    # TODO Rename this here and in `plot`
    def title_plot(self, arg0, arg1):
        plt.title(arg0)
        plt.grid(True)
        plt.subplot(3, 3, arg1)

    def _interpolate_airfoils(self):
        v = [
            list(
                zip(
                    [i for _ in self.airfoils[i][0]],
                    self.airfoils[i][0],
                    self.airfoils[i][1],
                )
            )
            for i in sorted(self.airfoils)
        ]
        nv = []
        for i in zip(*v):
            t, x, y = zip(*i)
            nx = np.interp(self.thickness[1], t, x)
            ny = np.interp(self.thickness[1], t, y)
            nv.append(list(zip(nx, ny)))

        sections = []
        for i in zip(*nv):
            x, y = list(zip(*i))
            sections.append(blade_section.section(x, y))

        return sections

    def _place_airfoils(self):
        self.sections = self._interpolate_airfoils()

        # build the blade up out of sections in two loops
        # first, scale and twist the section
        # a variable is created, analogous to focus, which represents the
        # fraction of the chord around which the twist is defined
        twist_center = 0.5

        for i in zip(self.x, self.chord[1], self.twist[1], self.sections):
            i[3].translate(-twist_center, 0.0, 0.0)
            i[3].twist(i[2])
            i[3].scale((i[1], i[1], 1.0))

        for i in self.sections:
            i.local_to_global()

        # then, translate the section coordinates to global
        for i in zip(self.sections, self.dx[1], self.dy[1], self.z[1]):
            i[0].translate(i[1], i[2], i[3])

    def export_variables(self, fname):
        var = {
            "dx": self.dx,
            # "dxf": self.dxf,
            "dy": self.dy,
            "z": self.z,
            "twist": self.twist,
            "chord": self.chord,
            "thickness": self.thickness,
            "absolute_thickness": self.absolute_thickness,
        }
        open(fname, "w").write(str(var))
        return var

    def dump(self, fname="__sections.txt", z_rotation=0.0):
        "dump to a sections list for use in FEA (with webs)"
        lst = [i.get_pointlist(z_rotation=z_rotation) for i in self.sections]
        if fname.endswith(".txt"):
            open(fname, "wb").write(str(lst).encode("utf-8"))
        elif fname.endswith(".pck"):
            pickle.dump(lst, open(fname, "wb"))

    def export_xfoil(self, prefix="airfoil_out/_xf"):
        """Export the sections to xfoil format

        parameters
        ----------
        prefix : str
            prefix to use for the filenames
        """
        for i in zip(self.sections, self.thickness[1], self.z[1]):
            nm = prefix + "_t_%.3f_r_%.3f" % (i[1], i[2])
            i[0].to_xfoil(nm.replace(".", "_") + ".dat")

    def mesh(self, fname=None):
        """Join up the sections to make a polydata object and save it to a file

        parameters
        ----------
        fname : str
            filename to save the polydata to.  If None, don't save to file.
        """
        n_points = self.np_chordwise
        points = []
        for i in self.sections:
            for j in range(i.polydata.GetNumberOfPoints()):
                pt = i.polydata.GetPoint(j)
                points.append([pt[0], pt[1], pt[2]])

        points = np.array(points)

        cells = []
        for i in range(1, len(self.sections)):
            s0 = range((i - 1) * n_points, i * n_points)
            s1 = range(i * n_points, (i + 1) * n_points)
            cells.extend(
                [4, s0[j], s1[j], s1[(j + 1) % n_points], s0[(j + 1) % n_points]]
                for j in range(n_points)
            )

        self.poly = pv.PolyData(points, np.hstack(cells))
        if fname != None:
            print("saving to", fname)
            self.poly.save(fname)



#! /usr/bin/env python3

import sys
import vtk
import os

import multiprocessing
from functools import partial
import argparse
import numpy as np
import pandas as pd
from numpy import array
from scipy.spatial import distance
import math
import copy
import pyvista as pv


def get_eids(sec, n):
    g = vtk.vtkIdList()
    sec.GetPointCells(n, g)
    return [g.GetId(i) for i in range(g.GetNumberOfIds())]


def get_nids(sec, e):
    l = vtk.vtkIdList()
    sec.GetCellPoints(e, l)
    return [l.GetId(i) for i in range(l.GetNumberOfIds())]


def join_up(nodes, epid, stacks, sec, poly):
    """
    take nodes where ply drops happen, link them up using vtkDelaunay2D (using
    scaled down thicknesses
    """
    cells = poly.GetPolys()
    mat = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")
    for inn in nodes:  # loop over all nodes that need joining up
        i, el = inn
        s0, s1 = np.array(stacks[el[0]]), np.array(stacks[el[1]])

        s0 /= np.linalg.norm(s0)
        s1 /= np.linalg.norm(s1)

        pnts = vtk.vtkPoints()
        for j in s0:
            pnts.InsertNextPoint(0, j, 0)
        for j in s1:
            pnts.InsertNextPoint(5, j, 0)

        pl = vtk.vtkPolyData()
        pl.SetPoints(pnts)
        d2 = vtk.vtkDelaunay2D()
        d2.SetInputData(pl)
        d2.Update()
        dd = d2.GetOutput()

        # new linking, fixes changes in ordering of elements
        x0 = 1
        ni0 = get_nids(sec, el[0])
        ni1 = get_nids(sec, el[1])
        if ni0[0] == ni1[1]:
            x0 = 0
        nmap = list(
            range(epid[el[0]] + x0, epid[el[0]] + x0 + (len(s0)) * 2, 2)
        ) + list(range(epid[el[1]] + 1 - x0, epid[el[1]] + 1 - x0 + (len(s1)) * 2, 2))
        for j in range(dd.GetNumberOfCells()):
            c = dd.GetCell(j)
            cells.InsertNextCell(4)
            for k in range(c.GetNumberOfPoints()):
                cells.InsertCellPoint(nmap[c.GetPointId(k)])
            cells.InsertCellPoint(nmap[c.GetPointId(k)])
            mat.InsertNextTuple1(-1)
            angle.InsertNextTuple1(0)


def align_normals(align):
    for i in range(align.GetNumberOfCells()):
        c = align.GetCell(i)
        p = c.GetPoints()
        if p.GetNumberOfPoints() > 2:
            normal = [0, 0, 0]
            vtk.vtkTriangle().ComputeNormal(
                p.GetPoint(0), p.GetPoint(1), p.GetPoint(2), normal
            )
            if normal[2] > 0:
                align.ReverseCell(i)


def web_link(web_links, epid, stacks, sec, poly):
    enormal = sec.GetCellData().GetArray("Normals")
    mat = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")
    cells = poly.GetPolys()

    for i in web_links:
        # print(i)
        to1 = i[2][0]  # connected element 1
        to2 = i[2][1]  # connected element 2
        s1 = stacks[to1]  # stack of element 1
        s2 = stacks[to2]  # stack of connected element 2
        s0 = stacks[i[0]]  # stack of web element
        n11 = epid[to1] + 2 * len(s1) - 1
        n12 = epid[to1] + 2 * len(s1) - 2
        n21 = epid[to2] + 2 * len(s2) - 1
        n22 = epid[to2] + 2 * len(s2) - 2

        n1, n2 = enormal.GetTuple3(i[0]), enormal.GetTuple3(to2)

        if max(n1[0] * n2[0], n1[1] * n2[1]) > 0.5:
            linkto = [j + epid[to2] for j in range(0, len(s2) * 2, 2)]
        else:
            linkto = [n11, n12, n21, n22]

        nds = linkto + [j + epid[i[0]] + i[3] for j in range(0, len(s0) * 2, 2)]

        pts = vtk.vtkPoints()

        p1 = poly.GetPoint(n11)
        p2 = poly.GetPoint(nds[4])

        if linkto[0] == n11:
            dx = -1 if p1[0] < p2[0] else 1
        else:
            dx = 0

        for j in nds[:4]:
            p = poly.GetPoint(j)
            pts.InsertNextPoint(p[0] + dx, p[1], 0)

        for j in nds[4:]:
            p = poly.GetPoint(j)
            pts.InsertNextPoint(p[0], p[1], 0)

        pol = vtk.vtkPolyData()
        pol.SetPoints(pts)
        d2 = vtk.vtkDelaunay2D()
        d2.SetInputData(pol)

        d2.Update()

        dd = d2.GetOutput()

        for j in range(dd.GetNumberOfCells()):
            c = dd.GetCell(j)
            cells.InsertNextCell(4)
            for k in range(c.GetNumberOfPoints()):
                cells.InsertCellPoint(nds[c.GetPointId(k)])

            cells.InsertCellPoint(nds[c.GetPointId(k)])
            mat.InsertNextTuple1(-1)
            angle.InsertNextTuple1(0)


def create_bondline(y, epid, stacks, sec, poly, bondline_material=20):
    lst = []

    bnds = sec.GetBounds()
    for i in range(sec.GetNumberOfPoints()):
        pnt = sec.GetPoint(i)
        if pnt[1] > bnds[3] - y:
            lst.extend(get_eids(sec, i))

    pts = vtk.vtkPoints()
    nds = []
    for j in set(lst):
        s = stacks[j]
        n1 = epid[j] + 2 * len(s) - 1
        n2 = epid[j] + 2 * len(s) - 2
        pts.InsertNextPoint(poly.GetPoint(n1))
        pts.InsertNextPoint(poly.GetPoint(n2))
        nds.extend((n1, n2))

    pol = vtk.vtkPolyData()
    pol.SetPoints(pts)
    d2 = vtk.vtkDelaunay2D()
    d2.SetInputData(pol)

    dd = d2.GetOutput()

    cells = poly.GetPolys()
    material = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")

    for j in range(dd.GetNumberOfCells()):
        c = dd.GetCell(j)
        cells.InsertNextCell(4)
        for k in range(c.GetNumberOfPoints()):
            cells.InsertCellPoint(nds[c.GetPointId(k)])
        cells.InsertCellPoint(nds[c.GetPointId(k)])
        material.InsertNextTuple1(bondline_material)
        angle.InsertNextTuple1(0)


def get_local_twist(r, var):
    return np.interp(r, var["z"][1] - np.min(var["z"][1]), var["twist"][1])


def get_local_chord(r, var):
    return np.interp(r, var["z"][1] - np.min(var["z"][1]), var["chord"][1])


def write_vtp(section, vtp):
    wr = vtk.vtkXMLPolyDataWriter()
    wr.SetFileName(vtp)
    wr.SetInputData(section)
    wr.Update()
    wr.Write()
    print(f"writing to {vtp}")


def get_avg_normal(normals):
    mean_normal = np.mean(normals, axis=0)
    mean_normal /= np.linalg.norm(mean_normal)
    return mean_normal


def get_node_avg_normal(connected_cell_normals):
    s = connected_cell_normals.shape
    if s[0] < 3:
        return get_avg_normal(connected_cell_normals), None
    corr = np.zeros((s[0], s[0]))
    for i in range(s[0]):
        for j in range(i + 1, s[0]):
            corr[i, j] = np.abs(
                np.dot(connected_cell_normals[i, :], connected_cell_normals[j, :])
            )
    if corr[0, 1] > np.mean([corr[0, 2], corr[1, 2]]):
        return get_avg_normal(connected_cell_normals[:2, :]), 2
    elif corr[0, 2] > np.mean([corr[0, 1], corr[1, 2]]):
        return get_avg_normal(connected_cell_normals[::2, :]), 1
    else:
        return get_avg_normal(connected_cell_normals[1:, :]), 0


def cut_blade(r, vtu, if_bondline=True, rotz=0, var=None, is2d=False, verbose=False):
    if var is None:
        var = {}
    print("# creating cross section mesh from %s at r=%.3f" % (vtu, r))
    workdir = os.path.dirname(vtu)
    # read in the mesh
    rd = pv.read(vtu)

    local_twist = get_local_twist(r, var)
    local_chord = get_local_chord(r, var)

    sec = rd.slice(normal=[0, 0, 1], origin=[0, 0, r])

    # if not is2d:
    #     # slice the mesh at some point
    #     cutter = vtk.vtkCutter()
    #     plane = vtk.vtkPlane()
    #     plane.SetOrigin(0, 0, r)
    #     plane.SetNormal(0, 0, 1)

    #     cutter.SetCutFunction(plane)
    #     cutter.SetInputData(msh)
    #     cutter.Update()
    #     sec = cutter.GetOutput()
    # else:
    #     sec = msh

    # first we rotate around the Z axis as if rotating the whole blade (compensate for pck angle)
    transform = vtk.vtkTransform()
    transform.RotateZ(rotz)
    transformfilter = vtk.vtkTransformFilter()
    transformfilter.SetTransform(transform)
    transformfilter.SetInputData(sec)
    transformfilter.Update()
    rotated_section = transformfilter.GetOutput()

    pts = np.array(
        [
            rotated_section.GetPoint(i)
            for i in range(rotated_section.GetNumberOfPoints())
        ]
    )
    minxy = np.amin(pts, axis=0)
    maxxy = np.amax(pts, axis=0)
    mid_position = 0.5 * (minxy + maxxy)

    bnds = rotated_section.GetBounds()

    table_out = pd.DataFrame(
        [
            [r]
            + list(mid_position)
            + [local_twist, local_chord]
            + [bnds[1] - bnds[0], bnds[3] - bnds[2]],
        ],
        columns=["r", "xavg", "yavg", "zavg", "twist_angle", "local_chord", "dx", "dy"],
    )

    # get a handle with plydata on points and cells
    ctp = vtk.vtkCellDataToPointData()
    ctp.PassCellDataOn()
    ctp.SetInputData(sec)
    ctp.Update()

    sec = ctp.GetOutput()

    # special case that deals with attachment of added webs that might link
    # crossed to elements (so the slice mesh is not linked up), this code links
    # up the slice by searching in negative y direction until the closest point
    # is no longer the unconnected point
    for i in range(sec.GetNumberOfPoints()):
        pb = sec.GetPoint(i)
        e0 = get_eids(sec, i)
        if len(e0) == 1:
            for j in np.linspace(0, 1, 100):
                p = sec.FindPoint(pb[0], pb[1] - j, pb[2])
                if p != i:
                    sec.GetPoints().SetPoint(i, sec.GetPoint(p))
                    break

    # clean up the mesh
    cln = vtk.vtkCleanPolyData()
    cln.SetInputData(sec)
    cln.Update()

    sec = cln.GetOutput()

    # get the max number of plies

    points = vtk.vtkPoints()
    cells = vtk.vtkCellArray()
    poly = vtk.vtkPolyData()
    poly.SetPoints(points)
    poly.SetPolys(cells)

    material = vtk.vtkIntArray()
    material.SetName("mat")
    poly.GetCellData().AddArray(material)
    angle = vtk.vtkFloatArray()
    angle.SetName("angle")
    poly.GetCellData().AddArray(angle)

    nply = sec.GetCellData().GetArray("n_plies")
    thickness = sec.GetCellData().GetArray("thickness")
    enormal = sec.GetCellData().GetArray("Normals")

    # gather cell-wise ply arrays
    plies = []
    for i in range(sec.GetCellData().GetNumberOfArrays()):
        name = sec.GetCellData().GetArrayName(i)
        if name.startswith("ply"):
            plies.append((name, sec.GetCellData().GetArray(i)))

    pid = 0
    epid = []
    join_nodes = []
    web_links = []
    stacks = []  # stack of plies
    mstacks = []  # material stack
    astacks = []  # angle stack

    # loop over all cells
    for i in range(sec.GetNumberOfCells()):
        # get the id, the nodes attached to the cell, and the elements attached to each node
        epid.append(pid)
        nn = get_nids(sec, i)
        e0 = get_eids(sec, nn[0])
        e1 = get_eids(sec, nn[1])

        e0b, e1b = copy.deepcopy(e0), copy.deepcopy(e1)

        # get the normal in node 0, and if there is a web connected, give back the id in e0 of the web element
        nr0, wn0 = get_node_avg_normal(np.array([enormal.GetTuple3(i) for i in e0]))
        # get the normal in node 1
        nr1, wn1 = get_node_avg_normal(np.array([enormal.GetTuple3(i) for i in e1]))

        if wn0 != None:
            del e0[wn0]
        if wn1 != None:
            del e1[wn1]

        x0, x1 = 0.0, 1.0
        if nply.GetTuple1(e0[0]) != nply.GetTuple1(e0[1]) or thickness.GetTuple1(
            e0[0]
        ) != thickness.GetTuple1(e0[1]):
            join_nodes.append((nn[0], e0))
            x0 = 0.1
        if nply.GetTuple1(e1[0]) != nply.GetTuple1(e1[1]) or thickness.GetTuple1(
            e1[0]
        ) != thickness.GetTuple1(e1[1]):
            x1 = 0.9

        if wn0 != None and i == e0b[wn0]:
            x0, x1 = 0, 1.0
            fact = 1.1
            cell_length = sec.GetCell(i).GetLength2() ** 0.5

            nr0 = np.array(enormal.GetTuple3(i))
            nr0[2] = 0.0
            nr0 /= np.linalg.norm(nr0)
            web_links.append([i, nn[0], e0b, 0])
            for jj in range(3):  # loop over the 3 elements attached to n0
                if wn0 != jj:  # if one isn't the current (web) element
                    th = thickness.GetTuple1(e0b[jj]) * 1e-3  # get the thickness
                    break
            x0 = min(fact * th / cell_length, 0.97)

        if wn1 != None and i == e1b[wn1]:
            x0, x1 = 0, 1.0
            fact = 1.1
            cell_length = sec.GetCell(i).GetLength2() ** 0.5

            nr1 = np.array(enormal.GetTuple3(i))
            nr1[2] = 0.0
            nr1 /= np.linalg.norm(nr1)
            web_links.append([i, nn[1], e1b, 1])
            th = thickness.GetTuple1(e1[0]) * 1e-3
            x1 = max(1.0 - fact * th / cell_length, 0.03)

        p0, p1 = zip(
            *[
                [(1.0 - x0) * ii[0] + x0 * ii[1], (1.0 - x1) * ii[0] + x1 * ii[1]]
                for ii in zip(sec.GetPoint(nn[0]), sec.GetPoint(nn[1]))
            ]
        )

        points.InsertNextPoint(p0), points.InsertNextPoint(p1)
        pid += 2
        stack = [0]
        mstack = []
        astack = []
        for j in plies:
            tup = j[1].GetTuple(i)
            if tup[0] != -1 and tup[1] > 0.01:  # filter out very thin plies (like the
                # ones used for puck postprocessing)
                # if the material and angle are the same as the last, don't put in a new ply,
                # but increase the previous ply thickness, this greatly reduces element count in e.g. sparcaps
                if mstack and (mstack[-1] == int(tup[0]) and astack[-1] == tup[2]):
                    stack[-1] += tup[1]
                    p0 = [k[0] + k[1] * tup[1] * 1e-3 for k in zip(p0, nr0)]
                    p1 = [k[0] + k[1] * tup[1] * 1e-3 for k in zip(p1, nr1)]
                    npts = points.GetNumberOfPoints()
                    points.SetPoint(npts - 2, tuple(p0))
                    points.SetPoint(npts - 1, tuple(p1))
                else:
                    stack.append(tup[1] + stack[-1])
                    mstack.append(int(tup[0]))
                    astack.append(tup[2])
                    p0 = [k[0] + k[1] * tup[1] * 1e-3 for k in zip(p0, nr0)]
                    p1 = [k[0] + k[1] * tup[1] * 1e-3 for k in zip(p1, nr1)]
                    points.InsertNextPoint(tuple(p0))
                    points.InsertNextPoint(tuple(p1))
                    pid += 2

        stacks.append(stack)
        mstacks.append(mstack)
        astacks.append(astack)

    # fill up the "normal" elements
    for i in range(sec.GetNumberOfCells()):
        p1 = epid[i]
        p2 = epid[i] + 1
        for j in range(len(mstacks[i])):
            cells.InsertNextCell(4)
            cells.InsertCellPoint(p1 + 2 * j)
            cells.InsertCellPoint(p1 + 2 * j + 2)
            cells.InsertCellPoint(p2 + 2 * j + 2)
            cells.InsertCellPoint(p2 + 2 * j)
            material.InsertNextTuple1(mstacks[i][j])
            angle.InsertNextTuple1(astacks[i][j])

    # join up the plydrop elements
    join_up(join_nodes, epid, stacks, sec, poly)

    # join up the webs with the shell
    web_link(web_links, epid, stacks, sec, poly)

    # create TE bondline
    if if_bondline:
        create_bondline(0.1, epid, stacks, sec, poly)

    if verbose:
        write_vtp(poly, os.path.join(workdir, "inter_%i.vtp" % (1e3 * r)))

    # clean up the polydata, i.e. remove duplicate points
    cln = vtk.vtkCleanPolyData()
    cln.SetInputData(poly)
    cln.Update()

    #  reapply the transform to the output mesh (run the mesh generator in original orientation)
    transform = vtk.vtkTransform()
    transform.RotateZ(rotz)
    transformfilter = vtk.vtkTransformFilter()
    transformfilter.SetTransform(transform)
    transformfilter.SetInputData(cln.GetOutput())
    transformfilter.Update()
    rotated_section = transformfilter.GetOutput()

    transform = vtk.vtkTransform()
    transform.Translate(-mid_position[0], -mid_position[1], -mid_position[2])
    transformfilter = vtk.vtkTransformFilter()

    transformfilter.SetTransform(transform)
    transformfilter.SetInputData(rotated_section)
    transformfilter.Update()
    zeroed_section = transformfilter.GetOutput()

    transform = vtk.vtkTransform()
    transform.RotateZ(local_twist)
    transformfilter = vtk.vtkTransformFilter()
    transformfilter.SetTransform(transform)

    transformfilter.SetInputData(zeroed_section)
    transformfilter.Update()
    detwisted_section = transformfilter.GetOutput()

    out = detwisted_section
    mat = out.GetCellData().GetArray("mat")
    ang = out.GetCellData().GetArray("angle")

    # get material and angle properties for the triangle elements
    for i in range(out.GetNumberOfCells()):
        c = out.GetCell(i)
        npts = c.GetNumberOfPoints()
        m = mat.GetTuple1(i)
        if npts == 3 and m == -1:
            g = c.GetPointIds()
            nb = []
            for j in [(0, 1), (1, 2), (2, 0)]:
                t = vtk.vtkIdList()
                t.InsertNextId(g.GetId(j[0]))
                t.InsertNextId(g.GetId(j[1]))

                idl = vtk.vtkIdList()
                out.GetCellNeighbors(i, t, idl)

                nb.extend(idl.GetId(k) for k in range(idl.GetNumberOfIds()))
            mat.SetComponent(i, 0, -1)
            ang.SetComponent(i, 0, 0)

            for j in nb:
                cp = out.GetCell(j).GetNumberOfPoints()
                if cp == 4:
                    mat.SetComponent(i, 0, mat.GetTuple1(j))
                    ang.SetComponent(i, 0, ang.GetTuple1(j))

    ang2 = vtk.vtkFloatArray()
    ang2.SetName("angle2")
    ang2.SetNumberOfTuples(out.GetNumberOfCells())
    angle_lookup = {}

    c = 1
    for i in range(out.GetNumberOfCells()):
        cl = out.GetCell(i)

        if cl.GetNumberOfPoints() > 2:
            [pt1, pt2] = [cl.GetEdge(1).GetPoints().GetPoint(j) for j in range(2)]

            vec = np.array([pt2[j] - pt1[j] for j in range(3)])
            vec /= np.linalg.norm(vec)
            dot = vtk.vtkMath.Dot([1, 0, 0], vec)

            angle = math.degrees(math.acos(dot))

            if pt2[1] < pt1[1]:
                angle *= -1

            if cl.GetNumberOfPoints() == 3:  # if
                for jj in range(3):
                    if cl.GetPointId(jj) in angle_lookup:
                        angle = angle_lookup[cl.GetPointId(jj)]
            else:
                angle_lookup[cl.GetPointId(0)] = angle

            ang2.SetComponent(i, 0, angle)

    out.GetCellData().AddArray(ang2)

    if verbose:
        write_vtp(out, os.path.join(workdir, "prerealign_%i.vtp" % (1e3 * r)))

    align_normals(out)

    of = os.path.join(workdir, "msec_%i.vtp" % (1e3 * r))
    wrt = vtk.vtkXMLPolyDataWriter()
    wrt.SetInputData(out)
    wrt.SetFileName(of)
    wrt.Write()
    print(f"# written vtk to {of}")
    table_out.to_csv(
        os.path.join(workdir, "section_location_%i.csv" % (1e3 * r)), index=False
    )


def cut_blade_parallel(
    vtu, rr, if_bondline, rotz, var, verbose=False, is2d=False, debug=False
):
    var = eval(open(var, "r").read())
    part = partial(
        cut_blade,
        vtu=vtu,
        if_bondline=if_bondline,
        rotz=rotz,
        var=var,
        is2d=is2d,
        verbose=verbose,
    )
    if debug == False:
        pool = multiprocessing.Pool()
        al = pool.map(part, rr)
        pool.close()
        pool.join()
    else:
        for i in rr:
            part(i)


def main():
    p = argparse.ArgumentParser(
        description="slice through the 3D FEA mesh (joined) to create 2D solid meshes for computing aeroelastic beam properties"
    )
    p.add_argument("--vtu", help="mesh input")
    p.add_argument("--r", default="[30]", help="radius array")
    p.add_argument("--bondline", default=False)
    p.add_argument("--rotz", type=float, default=0)
    p.add_argument("--var", type=str)
    p.add_argument("--verbose", action="store_true")
    p.add_argument("--debug", action="store_true")
    p.add_argument("--is2d", action="store_true")
    args = p.parse_args()

    rr = eval(args.r)
    cut_blade_parallel(
        args.vtu,
        rr,
        args.bondline,
        args.rotz,
        args.var,
        args.verbose,
        args.is2d,
        args.debug,
    )


if __name__ == "__main__":
    main()



import vtk


def intp_k(x, points, const=2, clamp=True, tension=-0.3, bias=0, continuity=0):
    """
    Kochanek spline, wrapping `vtk.vtkKochanekSpline()
    <http://www.vtk.org/doc/nightly/html/classvtkKochanekSpline.html>`_
    args:
        x (list): t coordinate list (in range=[0,1])
        points (list) : coordinate x,y,z points
        const (int): constraint type for left and right of spline
        clamp (bool): flag on whether spline is clamped
        tension (float): tension parameter
        bias (float): bias parameter value
        continuity (int) : continuity parameter value
    returns:
        x : t coordinate list
        o : evaluated coordinate list
    """
    sc = vtk.vtkKochanekSpline()
    sc.SetLeftConstraint(const)
    sc.SetRightConstraint(const)
    sc.SetDefaultTension(tension)
    sc.SetDefaultBias(bias)
    sc.SetDefaultContinuity(continuity)
    if clamp:
        sc.ClampValueOn()
    for i in points:
        sc.AddPoint(i[0], i[1])

    o = [sc.Evaluate(i) for i in x]
    return x, o


def intp_c(x, points, const=2, clamp=True):
    """
    Cardinal spline, wrapping `vtk.vtkCardinalSpline()
    <http://www.vtk.org/doc/nightly/html/classvtkCardinalSpline.html>`_
    args:
        x (list): t coordinate list (in range=[0,1])
        points (list) : coordinate x,y,z points
        const (int): constraint type for left and right of spline
        clamp (bool): flag on whether spline is clamped
    """
    try:
        sc = vtk.vtkCardinalSpline()
        sc.SetLeftConstraint(const)
        sc.SetRightConstraint(const)
        if clamp:
            sc.ClampValueOn()
        else:
            sc.ClampValueOff()
        for i in points:
            sc.AddPoint(i[0], i[1])

        o = [sc.Evaluate(i) for i in x]
        return x, o

    except Exception as e:
        print(points)
        exit(e)


def intp_sc(x, points):
    """
    SCurve spline based interpolation
    args:
        x (list) : t coordinate list
        points (list) : xyz coordinate input points
    returns:
        x (relative coordinate point list)
        o (xyz coordinate points list, resplined)
    """
    sc = vtk.vtkSCurveSpline()
    for i in points:
        sc.AddPoint(i[0], i[1])
    o = [sc.Evaluate(i) for i in x]
    return x, o



#! /usr/bin/env python

import fire
from b3p import (
    build_blade_geometry,
    build_blade_structure,
    build_plybook,
    drape_mesh,
    combine_meshes,
    mesh_2d,
    add_load_to_mesh,
    mesh2ccx,
    ccx2vtu,
    ccxpost,
    drape_summary,
    anba4_prep,
)
from utils import yml_portable
import os
import pickle
import numpy as np
import shutil

# from ruamel import yaml
import glob
import multiprocessing


class cli:
    """Fire command line interface for b3p"""

    def __init__(self, yml):
        self.dct = yml_portable.yaml_make_portable(yml, True)
        self.plybookname = "__plybook.pck"
        self.prefix = os.path.join(
            self.dct["general"]["workdir"], self.dct["general"]["prefix"]
        )

    def clean(self):
        """Clean up workdir"""
        if os.path.isdir(self.dct["general"]["workdir"]):
            shutil.rmtree(self.dct["general"]["workdir"])
            print(f"** Removing workdir {self.dct['general']['workdir']}")
        else:
            print(f"** Workdir {self.dct['general']['workdir']} does not exist")
        return self

    def __str__(self):
        return ""

    def geometry(self):
        """Build blade geometry based on yaml input file"""
        build_blade_geometry.build_blade_geometry(self.dct)
        # y = yaml.YAML()
        # y.dump(self.dct, open(f"{self.prefix}_portable.yml", "w"))
        yml_portable.save_yaml(f"{self.prefix}_portable.yml", self.dct)
        # yml_portable.save_yaml_portable(f"{self.prefix}_portable.yml")
        return self

    def mesh(self):
        """Build blade mesh based on yaml input file, requires blade geometry"""
        build_blade_structure.build_blade_structure(self.dct)
        return self

    def __plybook(self):
        """make plybook"""
        self.dct = build_plybook.expand_chamfered_cores(self.dct)
        build_plybook.lamplan2plies(self.dct, self.plybookname)
        return self

    def drape(self):
        """Assign plies in plybook to elements in the structural mesh"""
        self.__plybook()
        slb = self.dct["laminates"]["slabs"]
        used_grids = {slb[i]["grid"] for i in slb}

        pbook = os.path.join(self.dct["general"]["workdir"], self.plybookname)

        if os.path.exists(pbook):
            plybook = pickle.load(open(pbook, "rb"))
            meshes = []
            for grid in used_grids:
                out = f"{self.prefix}_{grid}_dr.vtu"
                drape_mesh.drape_mesh(f"{self.prefix}_{grid}.vtp", plybook, grid, out)
                meshes.append(out)
            combine_meshes.combine_meshes(meshes, f"{self.prefix}_joined.vtu")
        else:
            raise FileNotFoundError("Plybook not found")
        return self

    def mesh2d(self, z_start=0, z_end=100, nsec=100):
        """Create 2d meshes for calculation of 6x6 stiffness and matrices"""
        mesh_2d.cut_blade_parallel(
            f"{self.prefix}_joined.vtu",
            np.linspace(z_start, z_end, nsec).tolist(),
            if_bondline=False,
            rotz=0.0,
            var=f"{self.prefix}.var",
        )
        anba4_prep.anba4_prep(glob.glob(self.dct["general"]["workdir"] + "/msec*vtp"))
        return self

    def show(self):
        print(self)
        return self

    def ccxprep(
        self,
        merge_adjacent_layers=True,
        single_step=False,
        quadratic=True,
        force_isotropic=False,
        zeroangle=False,
        add_centers=False,
        export_plygroups=False,
        export_hyperworks=False,
        meshonly=False,
        buckling=False,
    ):
        print("** create ccx input file")
        mesh2ccx.mesh2ccx(
            f"{self.prefix}_joined.vtu",
            matmap=os.path.join(self.dct["general"]["workdir"], "material_map.json"),
            out=f"{self.prefix}_ccx.inp",
            merge_adjacent_layers=merge_adjacent_layers,
            single_step=single_step,
            quadratic=quadratic,
            force_isotropic=force_isotropic,
            zeroangle=zeroangle,
            add_centers=add_centers,
            export_plygroups=export_plygroups,
            buckling=buckling,
            meshonly=meshonly,
            export_hyperworks=export_hyperworks,
        )
        return self

    def ccxsolve(self, wildcard="", nproc=3):
        """Run ccx on all inp files in workdir that match wildcard"""

        inps = glob.glob(f"{self.prefix}*{wildcard}*inp")

        if inps == []:
            print(f"** No inps found matching {self.prefix}*{wildcard}*inp")
            return self

        inps_to_run = []
        for inp in inps:
            frd_file = inp.replace(".inp", ".frd")
            # check if frd exists
            if not os.path.exists(frd_file):
                inps_to_run.append(inp)
            else:
                # check if the frd is a complete file
                with open(frd_file, "rb") as f:
                    f.seek(-5, 2)
                    y = f.read()
                    if y != b"9999\n":
                        inps_to_run.append(inp)
                    else:
                        print(f"** Skipping {inp} because {frd_file} exists")

        print(f"running ccx on: {inps_to_run} using {wildcard}")
        p = multiprocessing.Pool(nproc)
        p.map(os.system, [f"ccx {inp.replace('.inp','')}" for inp in inps_to_run])
        p.close()
        return self

    def ccxpost(self, wildcard="", nbins=60):
        """Postprocess ccx results"""
        post = ccx2vtu.ccx2vtu(self.dct["general"]["workdir"], wildcard=wildcard)
        post.load_grids()
        post.tabulate(nbins)
        return self

    def ccxplot(self, wildcard="", plot3d=True, plot2d=True):
        plotter = ccxpost.plot_ccx(self.dct["general"]["workdir"], wildcard=wildcard)
        if plot3d:
            plotter.plot3d()
        if plot2d:
            plotter.plot2d(wildcard=wildcard)
        return self

    def build(self):
        """Build the whole model, geometry, mesh, drape"""
        self.geometry()
        self.mesh()
        self.drape()
        add_load_to_mesh.add_load_to_mesh(
            self.dct, f"{self.prefix}_joined.vtu", f"{self.prefix}_loads.png"
        )

        # self.mesh2d(z_start=0.0001, z_end=100, nsec=50)
        return self

    def mass(self):
        """Calculate mass of the blade, requires drape."""
        mass_table = drape_summary.drape_summary(f"{self.prefix}_joined.vtu")
        mass_table.to_csv(f"{self.prefix}_mass.csv")
        mass_table.replace(to_replace="_", value="\_", regex=True).to_latex(
            f"{self.prefix}_mass.tex", index=False
        )
        return self


def main():
    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(cli)


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import vtk
import numpy as np


def write_web(
    loc,
    normal,
    mesh,
    rootcut=0.0,
    tipcut=100.0,
    tip=0.0,
    zone="d_rel_dist_from_te",
):
    """
    Slices a plane through a mesh, and reports back a relative
    coordinate of top and bottom lines. This is used to represent geometrically straight entities
    in 3D as coordinates in local systems defined per section

    :param loc: location of the plane
    :param normal: normal of the plane
    :param mesh: mesh to slice
    :param rootcut: cut the root of the web
    :param tipcut: cut the tip of the web
    :param tip: tip of the blade
    """

    rd = vtk.vtkXMLPolyDataReader()
    rd.SetFileName(mesh)
    rd.Update()
    poly = rd.GetOutput()

    plane = vtk.vtkPlane()
    plane.SetOrigin(loc)
    plane.SetNormal(normal)

    clip = vtk.vtkCutter()
    clip.SetCutFunction(plane)
    clip.SetInputData(poly)

    clip.Update()
    out = clip.GetOutput()
    points = []

    c = out.GetPointData().GetArray("d_rel_dist_from_te")
    cc = out.GetPointData().GetArray("d_abs_dist_from_te")
    p = out.GetPointData().GetArray(zone)

    lw, ww = [], []
    lwp, wwp = [], []
    rta = []
    minr, maxr = 1000, -1000
    for i in range(out.GetNumberOfPoints()):
        if c.GetValue(i) > 0.5:
            lw.append((out.GetPoint(i)[2], p.GetValue(i)))
            lwp.append(out.GetPoint(i))
            rta.append((out.GetPoint(i)[2], cc.GetValue(i) / c.GetValue(i)))
        else:
            ww.append((out.GetPoint(i)[2], p.GetValue(i)))
            wwp.append(out.GetPoint(i))
        minr = min(minr, out.GetPoint(i)[2])
        maxr = max(maxr, out.GetPoint(i)[2])
        points.append(out.GetPoint(i))

    lw = list(zip(*lw))
    ww = list(zip(*ww))
    rt = list(zip(*rta))
    # interpolate the results
    r = np.linspace(minr, maxr, 400)
    lw1 = np.interp(r, lw[0], lw[1])
    ww1 = np.interp(r, ww[0], ww[1])
    rt1 = np.interp(r, rt[0], rt[1])

    out = []
    for i, rr in enumerate(r):
        if rr <= tipcut:
            lwl, wwl = lw1[i], ww1[i]

        out.append((rr, wwl, lwl, rt1[i]))

        if i > 0 and r[i - 1] < rootcut and rr > rootcut:
            for j in range(i):
                out[j][1] = wwl
                out[j][2] = lwl

    out = sorted(out)
    if tip > out[-1][0]:
        out.append((tip, out[-1][1], out[-1][2], out[-1][3]))

    return out


def build_webs(mesh, webs, prefix="__dum"):
    web_meshes = {}
    for i in webs:
        normal = (0, 1, 0)

        if "orientation" in webs[i]:
            normal = webs[i]["orientation"]

        fea_web = write_web(
            np.array(webs[i]["origin"]),
            normal,
            mesh,
            rootcut=webs[i]["z_start"],
            tipcut=webs[i]["z_follow_blade"],
            tip=webs[i]["z_end"],
        )
        web_meshes[i] = fea_web

    return web_meshes



#! /usr/bin/env python3

import numpy as np
import pyvista
from matplotlib import pyplot as plt
import pandas as pd


def compute_nodal_forces(nz, target_z, target_moment, fmult=1.0):
    assert fmult**2 == 1
    loaded = np.zeros_like(nz).astype(bool)
    force = np.zeros_like(nz)
    for i in reversed(list(zip(target_z, target_moment))):
        # compute the moment coming from the previously applied forces
        initial_moment = ((nz - i[0]) * fmult * force * loaded).sum()
        # get the nodes that are outboard from this point, but have not had a load applied yet
        rel_nodes = (nz > i[0]) & (loaded == 0)
        # compute distance for these nodes
        dz = (nz - i[0]) * rel_nodes
        # compute force needed to match moment curve
        fx_l = fmult * (i[1] - initial_moment) / dz.sum()
        # apply force
        force[np.where(rel_nodes)] = fx_l
        # register nodes as loaded
        loaded += rel_nodes

    zpl = np.unique(nz)
    moment = [((nz > j) * (nz - j) * force * fmult).sum() for j in zpl]

    return force, zpl, moment


def add_load_to_mesh(config, gridname, plotfile=None):
    grid = pyvista.UnstructuredGrid(gridname)

    if plotfile:
        fig, ax = plt.subplots(2, 1, figsize=(18, 13))

    lds = config["loads"]
    for i in lds:
        print(f"** loadcase {i}")
        # get applicable nodes
        for n, j in enumerate(config["loads"][i]["apply"]):
            key, [mn, mx] = j, config["loads"][i]["apply"][j]
            isloaded_thisrule = (grid.point_data[key] > mn) & (
                grid.point_data[key] < mx
            )
            if n == 0:
                is_loaded_node = isloaded_thisrule
            else:
                is_loaded_node = is_loaded_node & isloaded_thisrule

        loaded_node_ids = np.where(is_loaded_node)

        nz = grid.points[loaded_node_ids][:, 2]
        z = lds[i]["z"]
        mx = lds[i]["mx"]
        my = lds[i]["my"]

        assert len(z) == len(mx) == len(my)

        fx, zmy, bmy = compute_nodal_forces(nz, z, my, fmult=1.0)
        fy, zmx, bmx = compute_nodal_forces(nz, z, mx, fmult=-1.0)

        if plotfile:
            ax[0].plot(zmy, bmy, label=None)
            ax[0].plot(z, my, "o", label=f"my lc {i}")
            ax[0].plot(zmx, bmx, label=None)
            ax[0].plot(z, mx, "o", label=f"mx ls {i}")
            ax[0].legend(loc="best")
            ax[1].plot(nz, fx, label="fx, sum=%.2f" % fx.sum())
            ax[1].plot(nz, fy, label="fy, sum=%.2f" % fy.sum())
            ax[1].legend(loc="best")

        force_vector = np.zeros_like(grid.points)
        force_vector[loaded_node_ids, 0] = fx
        force_vector[loaded_node_ids, 1] = fy
        grid.point_data[f"lc_{i}"] = force_vector

    print(f"writing loadcases to grid {gridname}")
    grid.save(gridname)

    if plotfile:
        fig.savefig(plotfile)
        print(f"** written load plot to {plotfile}")
    return grid


if __name__ == "__main__":
    main()



#! /usr/bin/env python
import contextlib
import os
import numpy as np
import vtk


def load(fl, normalise=False):
    """
    load airfoil

    args:
        fl (str): filename

    kwargs:
        normalise (bool): flag determining whether the airfoil is normalised to
        unit length

    returns:
        [(x,y),...] airfoil point coordinates

    """
    d = []
    print(f"loading airfoil {fl}")
    for i in open(fl, "r"):
        with contextlib.suppress(Exception):
            xy = [float(j) for j in i.split()]
            if len(xy) in {2, 3}:
                d.append(xy)
    x, y = list(zip(*d))
    if normalise:
        mx = min(x)
        dx = max(x) - min(x)
        print("normalise factor %f" % dx)
        x = [(i - mx) / dx for i in x]
        y = [i / dx for i in y]

    return list(zip(x, y))


def optspace(n_points, base=0.2):
    """
    alternative to linspace for sampling that puts more points near the TE and
    LE of an airfoil
    """
    lep = 0.5
    x = np.linspace(0, 4.0 * np.pi, n_points)
    sp = 1.0 + base - np.cos(x)
    x1 = np.array([sum(sp[:i]) for i in range(len(x))])
    x1 = x1 / max(x1)
    return x1


def interp(x, points):
    """
    Interpolate airfoil points using 3D vtkParametricSpline

    args:

        x : List of points in range [0,1]

        points: List of 3d points [(x,y,z),...] to interpolate through
    """
    pnts = vtk.vtkPoints()
    for i in points:
        pnts.InsertNextPoint(i[0], i[1], 0 if len(i) == 2 else i[2])
    spline = vtk.vtkParametricSpline()
    spline.SetPoints(pnts)
    spline.SetLeftConstraint(3)
    spline.SetLeftValue(1.0)
    spline.SetRightConstraint(3)
    spline.SetRightValue(1.0)
    p, du = [0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0, 0]
    spline.DerivativesAvailableOn()
    out = []
    for i in x:
        p[0] = i
        u = [0, 0, 0]
        spline.Evaluate(p, u, du)
        out.append(u)
    return list(zip(*out))



from b3p import geom_utils
import vtk
import numpy as np
import math
import pyvista as pv

# from matplotlib import pyplot as plt
import copy


class section:
    """
    represents the geometry of a blade section

    args:
        r (list): list of radius position

        r_relative (list): list of relative radius positions

        points (list): list of points for this section

        min_te_thickness (float):  min TE thickness

    """

    def __init__(self, r, r_relative, points, min_te_thickness=0.002, open_te=True):
        self.r = r
        self.r_relative = r_relative
        if open_te:
            self.base_points = self._open_te(points, min_te_thickness)
        else:
            self.base_points = points
        self.te_thickness = min_te_thickness
        self.points = vtk.vtkPoints()
        for i in self.base_points:
            self.points.InsertNextPoint((i[0], i[1], r))
        self.poly = vtk.vtkPolyData()
        self.poly.SetPoints(self.points)

    def _open_te(self, points, te_thickness):
        """
        Open the trailing edge

        parameters:
            points (list): points
            te_thickness (float): trailing edge thickness
        """

        if len(points[0]) == 2:
            points = [list(i) + [0] for i in points]
        else:
            points = [list(i) for i in points]
        if (
            vtk.vtkMath.Distance2BetweenPoints(points[0], points[-1]) ** 0.5
            < te_thickness
        ):
            points[-1][1] -= (
                te_thickness
                - vtk.vtkMath.Distance2BetweenPoints(points[0], points[-1]) ** 0.5
            )

            for i in reversed(points):
                d = vtk.vtkMath.Distance2BetweenPoints(points[-1], i) ** 0.5
                if d >= 0.1:
                    break

                if i != points[-1]:
                    points.remove(i)
        return points

    def set_twist(self, twist):
        self.twist = twist
        self.poly.rotate_z(twist, inplace=True)

    def translate(self, dx, dy):
        self.poly.translate([dx, dy, 0.0], inplace=True)

    def scale(self, sx, sy):
        self.poly.scale([sx, sy, 1.0], inplace=True)

    def _create_evaluations(
        self, n_points, webs, mult=True, make_plots=False, panel_mesh_scale=None
    ):
        if panel_mesh_scale is None:
            panel_mesh_scale = []
        tol = 1e-9
        if webs == []:
            return np.linspace(0.0, 1.0, n_points)
        # compute where the web splits are on average (assuming they don't
        # cross)
        avspl = []
        for i in webs:
            avspl.extend(i.average_splits())

        # compute spacing between the splits
        intervals = sorted([0, 1] + list(avspl))
        isize = [intervals[i + 1] - intervals[i] for i in range(len(intervals) - 1)]

        for ps in panel_mesh_scale:
            if ps[0] < len(isize):
                isize[ps[0]] *= ps[1]

        # compute the number of cells there should be between each web split
        inum = [int(round(n_points * i / sum(isize))) for i in isize]

        if mult:
            isize[0] *= 2
            isize[-1] *= 2

        for _ in range(5):
            # consistent manner
            if (
                sum(inum) < n_points
            ):  # if not enough points, add one to the section with fewest points
                inum[inum.index(min(inum))] += 1
            elif (
                sum(inum) > n_points
            ):  # if too many, remove one from the section with most points
                inum[inum.index(max(inum))] -= 1

        splts = []
        for i in webs:
            splts.extend(i.splits(self.r, self.r_relative))

        real_intervals = sorted([0, 1] + list(splts))
        pnts = []
        for i in range(len(inum)):
            interval = np.linspace(
                real_intervals[i], real_intervals[i + 1], inum[i] + 1 * (i != 0)
            )
            pnts.extend(interval)

        return sorted(set(pnts))

    def respline(self, n_points, webs=None, added_datums=None, panel_mesh_scale=None):
        """
        re-evaluate the spline using n_points points, when there are webs they
        are put as hard points into the point list
        Add a whole bunch of related coordinates to the points, that can later
        be used to drape plies
        """
        if webs is None:
            webs = []
        if added_datums is None:
            added_datums = {}
        if panel_mesh_scale is None:
            panel_mesh_scale = []
        spline = vtk.vtkParametricSpline()
        spline.SetPoints(self.poly.GetPoints())
        spline.SetLeftConstraint(2)
        spline.SetRightConstraint(2)
        p, du = [0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0, 0]
        out = []
        last = []
        distance_along_airfoil = 0.0
        dist = []
        dist_from_te = []
        rel_dist_from_te = []
        x_abs, y_abs = [], []
        te = [
            (i[0] + i[1]) / 2.0
            for i in zip(
                self.poly.GetPoint(0),
                self.poly.GetPoint(self.poly.GetNumberOfPoints() - 1),
            )
        ]

        evaluations = self._create_evaluations(
            n_points, webs, panel_mesh_scale=panel_mesh_scale
        )
        for i in evaluations:
            p[0] = i
            u = [0, 0, 0]
            spline.Evaluate(p, u, du)
            out.append(u)
            if last == []:
                last = copy.deepcopy(u)
            distance_along_airfoil += geom_utils.distance(u, last)
            dist_from_te.append(geom_utils.distance(u, te))
            rel_dist_from_te.append(i)
            dist.append(distance_along_airfoil)
            x_abs.append(u[0])
            y_abs.append(u[1])
            last = copy.deepcopy(u)

        # compute a coordinate from the thickest point
        x, y, z = zip(*out)
        dist_miny = dist[y.index(min(y))]

        dist_sparcap = [i - dist_miny for i in dist]

        max_dist_from_te = max(dist_from_te)

        le_point = dist_from_te.index(max_dist_from_te)

        # chordline vector
        cline = [i[1] - i[0] for i in zip(out[le_point], te)]

        nor = [0, 0, 0]
        # get the in-plane normal to the chordline
        vtk.vtkMath.Cross(cline, [0, 0, -1], nor)
        chord = []
        for i in out:
            v1 = cline  # chord line
            v2 = nor  # normal to chord line
            v3 = [j[1] - j[0] for j in zip(te, i)]  # vector from TE to point

            chord.append(
                max(
                    0,
                    -(v3[0] / v1[0] - (v2[0] * v3[1]) / (v2[1] * v1[0]))
                    / (1.0 - (v2[0] * v1[1]) / (v2[1] * v1[0])),
                )
            )

        lechord = [-i + 1.0 for i in chord]
        lechord_absolute = [i * (cline[0] ** 2 + cline[1] ** 2) ** 0.5 for i in lechord]

        le_datum = [math.fabs(dist[le_point] - i) for i in dist]

        te_datum = [
            (dist[i] if i < le_point else max(dist) - dist[i]) for i in range(len(dist))
        ]

        # coordinates specific to the leading edge mould split line, defined as
        # the point where the y-coordinate is minimal
        sl_point = y_abs.index(min(y_abs))
        sl_dist = np.array(dist) - dist[sl_point]

        # suction and pressure sides have different false conditions to
        # distinguish them in case of centroid datasets
        suction_side = [1 if i < sl_point else -1 for i in range(len(dist))]
        pressure_side = [0 if i < sl_point else 1 for i in range(len(dist))]

        # add a coordinate that specifies the distance to the web attachment,
        # this is convenient for the specification of girders
        web_datums = {}
        for i in webs:
            splits = sorted(i.splits(self.r, self.r_relative))
            datum, datum_r = [], []
            for j in zip(rel_dist_from_te, pressure_side):
                datum.append(
                    (j[0] - splits[j[1]]) * distance_along_airfoil * (-1 if j[1] else 1)
                )
                datum_r.append((j[0] - splits[j[1]]) * (-1 if j[1] else 1))

            web_datums[f"d_{i.coordinate}"] = datum
            web_datums[f"d_{i.coordinate}_r"] = datum_r

        r = [self.r for _ in dist]
        # d_along_airfoil: distance from the TE to the point at hand along the
        # section
        # radius: radius position of the point
        # d_from_te: distance between current point and TE in a straight line
        # rel_dist_from_te: distance of point from TE scaled from 0 to 1
        # chord_length: chord length of the section the point is a part of
        # loc_le: d_along_airfoil coordinate of the leading edge point

        mdist = max(dist)
        datums = {
            "d_te": te_datum,
            "radius": r,
            "d_rel_dist_from_te": rel_dist_from_te,
            "d_abs_dist_from_te": dist,
            "d_abs_dist_from_bte": [-i + mdist for i in dist],
            "chord_length": [max_dist_from_te for _ in dist_from_te],
            "d_miny": dist_sparcap,
            "d_le": le_datum,
            "zone_ss": suction_side,
            "zone_ps": pressure_side,
            "d_te_r": [i / max(te_datum) for i in te_datum],
            "d_le_r": [i / max(le_datum) for i in le_datum],
            "d_chord": [1.0 - i for i in chord],
            "d_lechord": lechord,
            "d_lechord_abs": lechord_absolute,
            "d_x": x_abs,
            "d_y": y_abs,
            "d_sl": sl_dist,
            "d_sla": abs(sl_dist),
            "is_web": [0 for _ in r],
        }
        for i in web_datums:
            datums[i] = np.array(web_datums[i]).astype(np.float32)

        for i in added_datums.items():
            # print(self.r_relative, i[1][1], i[1][2])
            offs = np.interp(self.r_relative, i[1][1], i[1][2])
            datums[i[0]] = np.array(np.array(datums[i[1][0]]) + offs).astype(np.float32)

        return out, datums



#! /usr/bin/env python

from b3p import geometry_section
from b3p import geometry_blade_shape
from b3p import geometry_web
import pickle


def build_mesh(
    pckfile,
    radii,
    web_inputs,
    web_intersections,
    prefix,
    n_web_points=10,
    n_ch_points=120,
    outfile="out.vtp",
    added_datums=None,
    panel_mesh_scale=None,
):

    if added_datums is None:
        added_datums = {}
    if panel_mesh_scale is None:
        panel_mesh_scale = []
    sections = pickle.load(open(pckfile, "rb"))

    weblist = [
        geometry_web.web(
            points=web_intersections[i],
            web_root=web_inputs[i]["z_start"],
            web_tip=web_inputs[i]["z_end"],
            web_name=f"{prefix}_{i}.txt",
            coordinate=i,
            # normal=web_inputs[i]["normal"] if "normal" in web_inputs[i] else (0, 1, 0),
            flip_normal=(web_inputs[i]["origin"][1] > 0),
        )
        for i in web_inputs
    ]
    nsec = []
    z = [i[0][2] for i in sections]

    # loop over the sections and add them to a section list
    for i in sections:
        r = i[0][2] - min(z)
        r_rel = (r - min(z)) / (max(z) - min(z))
        sec = geometry_section.section(r, r_rel, i, open_te=False)
        nsec.append(sec)

    # create a blade loft from the sections
    blade = geometry_blade_shape.blade_shape(
        nsec,
        section_resolution=200,
        web_resolution=n_web_points,
        added_datums=added_datums,
    )

    for i in weblist:
        blade.set_web(i)

    blade.build_interpolated_sections(radii=radii, interpolation_type=2)

    # mesh with a given number of points around the circumference
    blade.mesh(n_ch_points, panel_mesh_scale=panel_mesh_scale)
    blade.write_mesh(outfile)
    print(f"** wrote mesh to {outfile}")

    return blade



#! /usr/bin/env python3
import fire
import pyvista as pv
import numpy as np
from matplotlib import pyplot as plt
import glob


def digitize_strain_distribution(z, strain, num_bins=100):
    # Bin the values in the first column into a specified number of bins
    zbin = np.linspace(z.min(), z.max(), num=num_bins + 1)
    bins = np.digitize(z, bins=zbin)

    # Find the maximum value in the second column for each bin
    max_vals = np.array(
        [np.max(strain[bins == bin_num]) for bin_num in range(1, num_bins + 1)]
    )
    min_vals = np.array(
        [np.min(strain[bins == bin_num]) for bin_num in range(1, num_bins + 1)]
    )

    return 0.5 * (zbin[:-1] + zbin[1:]), min_vals, max_vals


class plot_ccx:
    def __init__(self, wdir):
        self.meshes = glob.glob(wdir + "/*ccx*.vtu")

    def plot3d(self):
        for i in self.meshes:
            self.__plot3d(pv.read(i), output_path=i.replace(".vtu", "_3d.png"))
        return self

    def __plot3d(self, mesh, output_path):

        # Deform the mesh using the ChannelDisplacement filter
        deformed_mesh = mesh.warp_by_vector("disp", factor=1)

        # Color the mesh using the third component of point data strain
        deformed_mesh.point_data["strain"] = mesh.point_data["strain"][:, 2]
        deformed_mesh.set_active_scalars("strain")

        deformed_mesh.rotate_y(-110, inplace=True)

        deformed_mesh.rotate_z(0, inplace=True)

        # Create a plot in isometric view
        plotter = pv.Plotter(off_screen=True, window_size=[1920, 1080])
        plotter.add_mesh(
            deformed_mesh,
            cmap="coolwarm",
            clim=[-5e-3, 5e-3],
            scalar_bar_args={
                "position_x": 0.1,
                "position_y": 0.05,
                "width": 0.8,
                "height": 0.06,
                "color": "black",
                "n_labels": 2,
                "title_font_size": 22,
            },
        )
        plotter.view_isometric()
        plotter.camera.parallel_projection = True
        plotter.camera.zoom(2.0)

        # chart = pv.Chart2D(x_label="z (m)", y_label="strain (m/m)")

        # Save the plot to a high-resolution image
        plotter.screenshot(
            output_path,
            transparent_background=True,
            return_img=False,
        )
        print(f"Saved {output_path}")

    def __str__(self):
        return ""

    def plot2d(self):
        for i in self.meshes:
            self.__plot2d(pv.read(i), output_path=i.replace(".vtu", "_2d.png"))
        return self

    def __plot2d(self, mesh, output_path):
        """"""
        z, ymin, ymax = digitize_strain_distribution(
            mesh.points[:, 2], mesh.point_data["strain"][:, 2]
        )

        fig, ax = plt.subplots(figsize=(12, 8))
        ax.plot(z, ymin, color="blue", label="compression")
        ax.plot(z, ymax, color="red", label="tension")

        ax.set_xlabel("z-coordinate")
        ax.set_ylabel("strain")
        ax.legend()
        ax.set_ylim(-7e-3, 7e-3)
        ax.grid(True)
        fig.savefig(output_path)
        print(f"Saved {output_path}")


# def main():
#     deform_and_color_mesh

if __name__ == "__main__":
    fire.Fire(fea_results)



#! /usr/bin/env python3

import pyvista as pv
import numpy as np
import vtk
import fire
import time
import json
import os
import yaml
import pandas as pd


def zero_midside_loads(mesh):
    if mesh.celltypes[0] == 23:
        conn = mesh.cell_connectivity.reshape(
            (
                mesh.GetNumberOfCells(),
                int(mesh.cell_connectivity.shape[0] / mesh.GetNumberOfCells()),
            )
        )
        midsides = conn[:, 4:8].flatten()
        for i in mesh.point_data:
            if i.startswith("lc_"):
                mesh.point_data[i][midsides] *= 0.0
    return mesh


def make_shell_section(elem_id, plyarray, merge_adjacent_plies=True, zero_angle=True):
    plies = []
    filtered_plyarray = plyarray[plyarray[:, 1] > 1e-6]

    for j in filtered_plyarray:
        if plies and plies[-1][1] == j[0] and merge_adjacent_plies:
            plies[-1][0] += j[1] * 1e-3
        else:
            plies.append([j[1] * 1e-3, j[0]])

    if zero_angle:
        section_string = "".join("%f,,m%i,0\n" % tuple(i) for i in plies)
    else:
        section_string = "".join(
            "%f,,m%i,or%i\n" % tuple(i + [elem_id + 1]) for i in plies
        )

    return len(plies), section_string


def material_db_to_ccx(materials, matmap=None, force_iso=False):
    """find the material db and write properties to a ccx block"""
    # 2 files are relevant, a material map that maps the material ID (integer)
    # in the VTK file to a key (string) in the material database and the material
    # database itself
    mat_db = None
    if os.path.isfile(matmap):  # check if the material map file is there
        gdir = os.path.dirname(matmap)
        mm = json.load(open(matmap, "r"))
        if "matdb" in mm:  # check if the material map file points to a material db
            mat_db = yaml.load(
                open(os.path.join(gdir, mm["matdb"])), Loader=yaml.CLoader
            )
        else:
            exit(
                "material map available, but no link to material db, need matdb definition to do FEA"
            )
    else:
        exit("no material map defined")

    mm_inv = {v: k for k, v in mm.items()}

    matblock = ""
    for i in materials:
        if i > 1e-6:
            material_properties = mat_db[mm_inv[int(i)]]

            # print(material_properties)
            matblock += (
                f'** material: {mm_inv[int(i)]} {i} {material_properties["name"]}\n'
            )
            # matblock += f"** {str(material_properties)}\n"

            if (
                # "vf" in material_properties
                # or
                "C" in material_properties
                and not force_iso
            ):
                print(material_properties["name"], "is assumed to be orthotropic")
                C = np.array(material_properties["C"])
                # https://github.com/rsmith-nl/lamprop/blob/410ebfef2e14d7cc2988489ca2c31103056da38f/lp/text.py#L96
                # https://web.mit.edu/calculix_v2.7/CalculiX/ccx_2.7/doc/ccx/node193.html
                matblock += "** orthotropic material\n"
                matblock += "*material,name=m%i\n*elastic,type=ortho\n" % i
                D = C
                D[0, 3] = C[0, 5]
                D[0, 5] = C[0, 3]
                D[1, 3] = C[1, 5]
                D[1, 5] = C[1, 3]
                D[2, 3] = C[2, 5]
                D[2, 5] = C[2, 3]
                D[3, 3] = C[5, 5]
                D[5, 5] = C[3, 3]
                # D *= 1e6
                matblock += (
                    f"{D[0,0]:.4g},{D[0,1]:.4g},{D[1,1]:.4g},"
                    + f"{D[0,2]:.4g},{D[1,2]:.4g},{D[2,2]:.4g},"
                    + f"{D[3,3]:.4g},{D[4,4]:.4g},\n"
                    + f"{D[5,5]:.4g},293\n"
                )
            elif "e11" in material_properties and not force_iso:
                print(material_properties["name"], "has engineering constants")

                matblock += "** orthotropic material\n"
                matblock += (
                    "*material,name=m%i\n*elastic,type=engineering constants\n" % i
                )
                matblock += (
                    f"{material_properties['e11']:.4g},{material_properties['e22']:.4g},{material_properties['e33']:.4g},"
                    + f"{material_properties['nu12']:.4g},{material_properties['nu31']:.4g},{material_properties['nu23']:.4g},"
                    + f"{material_properties['g12']:.4g},{material_properties['g31']:.4g},\n"
                    + f"{material_properties['g23']:.4g},293\n"
                )

            else:
                print(material_properties["name"], "is assumed to be isotropic")
                # print(material_properties)
                nu = min(
                    0.45,
                    max(
                        0.1,
                        float(material_properties["nu"])
                        if "nu" in material_properties
                        else material_properties["nu12"],
                    ),
                )
                E = float(
                    material_properties["Ex"]
                    if "Ex" in material_properties
                    else material_properties["E"]
                )
                matblock += "** isotropic material\n"
                matblock += "*material,name=m%i\n*elastic,type=iso\n" % i
                matblock += f"{E:.4g},{nu:.4g},293\n"

    return matblock


def format_eset(name, eids):
    out = f"*elset,elset={name}\n"
    for i in range(len(eids)):
        out += f"{eids[i]}"
        out += "\n" if (i % 16 == 15) else ","
    if out[-1] == ",":
        out = out[:-1] + "\n"
    return out


def compute_ply_groups(grid, prefix):
    gr = ""
    out = []
    n = 1
    for i in grid.cell_data:
        if i.startswith(prefix):
            thickness = grid.cell_data[i][:, 1].max()
            material = grid.cell_data[i][:, 0].max()
            out.append(
                {
                    "Ply Name": i,
                    "Ply ID": n,
                    "Card Image": "PLY",
                    "Mat Name": f"m{int(material)}",
                    "Thickness": thickness,
                    "Orientation": 0,
                    "Output Results": "yes",
                    "TMANU": "",
                    "DRAPE_ID": 0,
                    "ESID": n,
                }
            )
            eids = np.where(grid.cell_data[i][:, 1] > 0)[0] + 1
            gr += format_eset(i, eids)
            n += 1
    return gr, pd.DataFrame(out)


def compute_slab_groups(grid, prefix):
    gr = ""
    for i in grid.cell_data:
        if i.startswith(prefix):
            eids = np.where(grid.cell_data[i] > 0)[0] + 1
            gr += format_eset(i, eids)
    return gr


def nodebuffer(grid):
    nodes = np.column_stack((np.arange(1, len(grid.points) + 1), grid.points))
    print(f"** exporting {len(nodes)} nodes")
    return (
        "*node,nset=nall\n"
        + "\n".join([f"{int(n)},{x:f},{y:f},{z:f}" for n, x, y, z in nodes])
        + "\n"
    )


def element_buffer(grid, quadratic=True):
    conn = grid.cell_connectivity.reshape((grid.GetNumberOfCells(), -1)) + 1
    eltype = "s8r" if quadratic else "s4"
    buf = ""
    for n, i in enumerate(conn):
        buf += f"*element,type={eltype},elset=e{n+1}\n"
        buf += f"{n+1},{','.join(map(str, i))}\n"

    return buf


def orientation_buffer(grid, add_centers=False):
    buf = ""
    # write orientation TODO match with element orientation, for now just align with z-axis
    x_dirs = grid.cell_data["x_dir"]
    y_dirs = grid.cell_data["y_dir"]
    centers = grid.cell_data["centers"]
    num_cells = grid.GetNumberOfCells()

    for n in range(num_cells):
        xdir, ydir = x_dirs[n], y_dirs[n]
        center = centers[n]
        buf += "*orientation,name=or%i,system=rectangular\n" % (n + 1)
        if add_centers:
            coords = np.concatenate([xdir + center, ydir + center, center], axis=0)
            buf += ",".join(format(k, ".4g") for k in coords.tolist()) + "\n3,0\n"
        else:
            coords = np.concatenate([xdir, ydir], axis=0)
            buf += ",".join(format(k, ".4g") for k in coords.tolist()) + "\n"

    return buf


def get_loadcases(mesh, multiplier=1.0, buckling=False):
    loadcases = {}

    for i in mesh.point_data:
        if i.startswith("lc_"):
            print(f"loadcase {i}")
            # forces are interpolated to midside nodes, causing the sum of forces to be off,
            # compute a multiplier from the sum of the forces in the linear model here
            multiplier = 1.0  # TODO fix for quadratic meshes # mesh.point_data[i].sum() / mesh.point_data[i].sum()
            if buckling:
                lbuf = f"** {i}\n*step\n*buckle\n5\n*cload\n"
            else:
                lbuf = f"** {i}\n*step\n*static\n*cload\n"

            ld = mesh.point_data[i] * multiplier
            for n, j in enumerate(ld):
                if j[0] ** 2 > 1e-8:
                    lbuf += "%i,1,%f\n" % (n + 1, j[0])
                if j[1] ** 2 > 1e-8:
                    lbuf += "%i,2,%f\n" % (n + 1, j[1])

            lbuf += "*node file,output=3d\nU,RF\n*EL FILE\nE\n*node print,nset=root,totals=yes\nrf\n*end step\n"
            # \n*node print,nset=nall\nrf
            # if buckling:
            #     lbuf += (
            #         "*step\n*buckle\n5\n*el file\n*node file,output=3d\nu\n*end step\n"
            #     )

            loadcases[i] = lbuf

    return loadcases


def root_clamp(mesh):
    root = np.where(mesh.points[:, 2] == mesh.points[:, 2].min())
    lbuf = "*nset, nset=root\n"
    for n, j in enumerate(root[0]):
        lbuf += "%i" % (j + 1)
        if n % 16 == 0:
            lbuf += "\n"
        else:
            lbuf += ","
    lbuf += "\n"
    lbuf += "*boundary,op=new\nroot,1,3\n"
    # for j in root[0]:
    #     lbuf += "%i,1,3\n" % (j + 1)
    return lbuf


def mesh2ccx(
    vtu,
    out="test.inp",
    matmap="temp/material_map.json",
    merge_adjacent_layers=False,
    zeroangle=False,
    single_step=False,
    quadratic=False,
    add_centers=False,
    force_isotropic=False,
    export_hyperworks=False,
    export_plygroups=False,
    buckling=False,
    meshonly=False,
):
    """
    Export a grid to ccx input file

    :param  grid (_type_): Grid file (vtu)
    :param out (str, optional): Output file. Defaults to "test.inp".
    :param matmap (str, optional): material map file. Defaults to "temp/material_map.json".
    :param merge_adjacent_layers (bool, optional): _description_. Defaults to False.
    :param zeroangle (bool, optional): _description_. Defaults to False.
    :param single_step (bool, optional): _description_. Defaults to False.
    :param quadratic (bool, optional): _description_. Defaults to False.
    :param add_centers (bool, optional): _description_. Defaults to False.
    :param force_isotropic (bool, optional): _description_. Defaults to False.
    :param export_hyperworks (bool, optional): _description_. Defaults to False.
    :param export_plygroups (bool, optional): _description_. Defaults to False.
    :param buckling (str, optional): _description_. Defaults to "static".
    :param meshonly (bool, optional): _description_. Defaults to False.
    :return: _description_
    """
    grid = pv.read(vtu)
    gr = grid.threshold(value=(1e-6, 1e9), scalars="thickness")
    gr.cell_data["centers"] = gr.cell_centers().points

    print(f"** Exporting {gr.GetNumberOfCells()} elements")
    if quadratic:
        lf = vtk.vtkLinearToQuadraticCellsFilter()
        lf.SetInputData(gr)
        lf.Update()
        quad = lf.GetOutput()
        mesh = zero_midside_loads(pv.UnstructuredGrid(quad))
        mesh.save(vtu.replace(".vtu", "_quad.vtu"))
    else:
        mesh = pv.UnstructuredGrid(gr)

    buf = nodebuffer(mesh)

    if quadratic:
        buf += element_buffer(mesh, quadratic=True)
    else:
        buf += element_buffer(mesh, quadratic=False)

    if meshonly:
        of = out.replace(".inp", "_meshonly.inp")
        open(of, "w").write(buf)
        print(f"** written to {of}")
        return

    buf += "*elset,elset=Eall,GENERATE\n%i,%i\n" % (1, mesh.GetNumberOfCells())

    print(f"** Export plygroups {export_plygroups}")
    if export_plygroups:
        plygroups, df = compute_ply_groups(mesh, "ply_")

        slabgroups = compute_slab_groups(mesh, "slab_thickness_")

        buf += plygroups + slabgroups

    plykeys = [i for i in mesh.cell_data if i.startswith("ply_")]

    plydat = np.stack([mesh.cell_data[i] for i in plykeys])

    # get all materials of all plies
    materials = np.unique(plydat[:, :, 0])

    buf += orientation_buffer(mesh, add_centers)

    print("** made orientation buffer")
    matblock = material_db_to_ccx(materials, matmap=matmap, force_iso=force_isotropic)

    buf += matblock

    tic = time.perf_counter()

    blx = [
        make_shell_section(i, plydat[:, i, :], merge_adjacent_layers, zeroangle)
        for i in range(plydat.shape[1])
    ]
    print("** made shell sections")

    nplies = np.array([i[0] for i in blx])

    nplmax = nplies.max()
    npxid = np.where(nplies == nplmax)[0]
    # print(npxid)
    print(f"max number of plies: {nplmax}")
    print(f"associated stack \n{ blx[npxid[0]][1]}")

    toc = time.perf_counter()
    print("** time spent creating shell sections %f" % (toc - tic))
    # sys.stdout.flush()
    comps = "".join(
        f"*shell section,composite,elset=e{n+1},offset=-.5"
        + (f",orientation=or{n+1}\n" if zeroangle else "\n")
        + i[1]
        for n, i in enumerate(blx)
    )
    buf += comps
    buf += root_clamp(mesh)

    loadcases = get_loadcases(mesh, buckling=buckling)

    # write a full ccx file for each loadcase, assuming parallel execution
    if single_step:
        output = buf + "".join(loadcases.values())
        open(out, "w").write(output)
        print(f"** written ccx input file with all loadcases to {out}")
        otb = out.replace(".inp", ".csv")
    else:
        for i in loadcases:
            output = buf + loadcases[i]
            of = out.replace(".inp", f"_{i}.inp")
            open(of, "w").write(output)
            print(f"** written ccx input file to {of}")

    if export_hyperworks:
        df.to_csv(otb, index=False)
        print(f"** written plybook to hyperworks table {otb}")


def main():
    fire.Fire(mesh2ccx)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import pandas as pd
import pyvista
from io import StringIO
import multiprocessing
import numpy as np
import vtk
import fire
import time

import regex


def findall(p, s):
    return [m.start() for m in regex.finditer(p, s, overlapped=True)]


def load_frd(inp):
    if inp.find("PSTEP") != -1:
        start_step_line = inp.find("100CL")
        end_step_line = inp.find("\n", start_step_line)
        _, solution_type, timestep = inp[start_step_line:end_step_line].split()[:3]

    if inp.find("CalculiX") != -1:  # read the nodes
        out = pd.read_fwf(
            StringIO(inp[inp.find("\n -1") :]),
            skiprows=1,
            names=["n", "x", "y", "z"],
            widths=[3, 10, 12, 12, 12],
            index_col=1,
        )
        return ("nodes", out.reset_index())
    elif inp.find("3C") > 1:  # elements
        e1 = pd.read_csv(
            StringIO(inp[inp.find("\n -1") :]),
            sep="\s+",
            skiprows=lambda x: (x % 3) != 0,
            names=range(11),
        )
        e2 = pd.read_csv(
            StringIO(inp[inp.find("\n -1") :]),
            sep="\s+",
            skiprows=lambda x: (x % 3) != 2,
            names=range(11),
        )
        return (
            "elements",
            pd.concat([e2.iloc[:, 1:], e1.iloc[:, 1:]], axis=1),
        )
    elif inp.find("DISP") != -1:  # displacements
        out = pd.read_fwf(
            StringIO(inp[inp.find("\n -1") :]),
            skiprows=1,
            names=["n", "d1", "d2", "d3"],
            widths=[3, 10, 12, 12, 12],
            index_col=1,
        ).astype("float32")
        return (f"disp_{timestep}", out)
    elif inp.find("FORC") != -1:  # forces
        out = pd.read_fwf(
            StringIO(inp[inp.find("\n -1") :]),
            skiprows=1,
            names=["n", "f1", "f2", "f3"],
            widths=[3, 10, 12, 12, 12],
            index_col=1,
        ).astype("float32")
        return (f"force_{timestep}", out)
    elif inp.find("STRESS") != -1:  # stresses
        out = pd.read_fwf(
            StringIO(inp[inp.find("\n -1") :]),
            skiprows=1,
            names=["n", "sxx", "syy", "szz", "sxy", "syz", "szx"],
            widths=[3, 10, 12, 12, 12, 12, 12, 12],
            index_col=1,
        ).astype("float32")
        return (f"stress_{timestep}", out)
    elif inp.find("STRAIN") != -1:  # strains
        out = pd.read_fwf(
            StringIO(inp[inp.find("\n -1") :]),
            skiprows=1,
            names=["n", "exx", "eyy", "ezz", "exy", "eyz", "ezx"],
            widths=[3, 10, 12, 12, 12, 12, 12, 12],
            index_col=1,
        ).astype("float32")
        return (f"strain_{timestep}", out)

    return ("other", None)


def frd2vtu(frd, output=None, multi=False):
    if not output:
        output = frd.replace(".frd", ".vtu")

    print("running frd2vtu on file: ", frd)
    tic = time.perf_counter()
    x = open(frd, "r").read()
    # find all block endings
    min3 = [0] + findall("\n -3", x)

    # pass the blocks to a process each
    if multi:
        p = multiprocessing.Pool()
        o = p.map(load_frd, [x[i[0] : i[1]] for i in zip(min3, min3[1:])])
    else:
        o = [load_frd(x[i[0] : i[1]]) for i in zip(min3, min3[1:])]

    o = dict(o)

    # create a map between ccx node id and vtk
    idmap = pd.DataFrame(o["nodes"].index, o["nodes"]["index"])

    # create new connectivity table
    el = o["elements"]
    # hex elements
    nna = el.isna().sum(axis=1).values
    hx = el.loc[nna == 0, :].values.astype(int)
    hx = np.hstack([hx[:, :12], hx[:, 16:], hx[:, 12:16]])

    # wedge elements
    wd = el.loc[nna == 5, :].values[:, :-5].astype(int)
    wd = np.hstack([wd[:, :9], wd[:, 12:], wd[:, 9:12]])

    hshape = hx.shape
    hx = idmap.loc[hx.flatten()].values.reshape(hshape)

    wshape = wd.shape
    wd = idmap.loc[wd.flatten()].values.reshape(wshape)

    # use the pyvista dict mesh build interface
    ogrid = pyvista.UnstructuredGrid(
        {vtk.VTK_QUADRATIC_HEXAHEDRON: hx, vtk.VTK_QUADRATIC_WEDGE: wd},
        o["nodes"][["x", "y", "z"]].values,
    )

    for i in o:
        if any(
            i.startswith(prefix) for prefix in ["disp", "force", "stress", "strain"]
        ):
            ogrid.point_data[i] = o[i].values[:, 1:]

    # add mises strain and stress

    # for s in [("stress", "s"), ("strain", "e")]:
    #     ss = o[s[0]]
    #     xx, yy, zz, xy, yz, zx = [
    #         ss[f"{s[1]}{i}"] for i in ["xx", "yy", "zz", "xy", "yz", "zx"]
    #     ]
    #     ogrid.point_data[f"mises_{s[0]}"] = (1.0 / 2.0**0.5) * (
    #         (xx - yy) ** 2
    #         + (yy - zz) ** 2
    #         + (zz - xx) ** 2
    #         + 6.0 * (xy**2 + yz**2 + zx**2)
    #     ) ** 0.5

    ogrid.save(output)

    print(f"** written to {output}")
    toc = time.perf_counter()
    print("time for translation %fs" % (toc - tic))
    return ogrid


if __name__ == "__main__":
    fire.Fire(frd2vtu)



#! /usr/bin/env python

import numpy as np
import re
import fire
import os

# import yaml
from ruamel.yaml import YAML


# def deep_convert_to_float(data):
#     """
#     Recursively convert all string representations of numbers in a nested data structure
#     (including dictionaries and lists) to floats.
#     """
#     if isinstance(data, dict):
#         return {k: deep_convert_to_float(v) for k, v in data.items()}
#     elif isinstance(data, list):
#         return [deep_convert_to_float(item) for item in data]
#     elif isinstance(data, str):
#         try:
#             return float(data)
#         except ValueError:
#             return data  # Return the original string if conversion is not possible
#     else:
#         return data  # Return the data as is if it's not a list, dict, or string


# # Define a custom representer for lists to make them inline
# def custom_list_representer(dumper, data):
#     return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)


# # Register the custom list representer with the PyYAML dumper
# yaml.add_representer(list, custom_list_representer)


def load_airfoil(af):
    """Load an airfoil from an xfoil formated text file.
    Check if the first line are coordinates or airfoil name."""
    with open(af, "r") as f:
        name = f.readline().strip()
        floats = re.findall(r"\d+\.\d+", name)
        if len(floats) == 2:
            offset = 0
            name = af.split("/")[-1].split(".")[0]
        elif name == "":
            offset = 1
            name = af.split("/")[-1].split(".")[0]
        else:
            offset = 1

    return name, np.loadtxt(af, skiprows=offset).tolist()


def load_airfoils(afs, prefix=""):
    """Load a list of airfoils from a dictionary of xfoil formated text files."""
    dct = {}
    for i in afs:
        if "xy" in afs[i]:
            return afs
        name, xy = load_airfoil(os.path.join(prefix, afs[i]))
        dct[i] = {"xy": xy}
        dct[i]["name"] = name
        dct[i]["path"] = afs[i]
        print(f"\t** imported airfoil {name} at thickness {i}")

    return dct


def yaml_make_portable(yaml_file, safe=False):
    """Import a yaml file and all linked files, and write to a _portable version of itself.

    :param yaml_file: path to yaml file
    :param safe: if True, use safe loader and basic data types (dict, list, str, int, float)
    :return: None"""
    print(f"** Loading yaml file {yaml_file} and loading linked files")
    prefix = os.path.dirname(yaml_file)

    yaml = YAML()  # typ="safe")
    # typ="safe")  # YAML(typ="safe") if safe else YAML(typ="rt")
    d = yaml.load(open(yaml_file, "r"))
    # yaml.load(open(yaml_file, "r"), Loader=yaml.CLoader)

    d["aero"]["airfoils"] = load_airfoils(d["aero"]["airfoils"], prefix=prefix)

    if type(d["materials"]) == str:
        print(f'\t** loading materials from: {d["materials"]}')
        d["materials"] = yaml.load(open(os.path.join(prefix, d["materials"]), "r"))

    if d["general"]["workdir"].find("portable") == -1:
        d["general"]["workdir"] += "_portable"
    return d  # deep_convert_to_float(d)


# Alternatively, define a custom representation for lists to ensure all are inline
def represent_list(dumper, data):
    return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)


def save_yaml(of, dct):
    yaml = YAML()
    yaml.default_flow_style = True
    yaml.representer.add_representer(list, represent_list)
    with open(of, "w") as f:
        yaml.dump(dct, f)
        print(f"\twritten to: {of}")


def save_yaml_portable(yaml_file):
    d = yaml_make_portable(yaml_file)
    of = yaml_file.replace(".yml", "_portable.yml").replace(".yaml", "_portable.yml")
    save_yaml(of, d)


def main():
    fire.Fire(save_yaml_portable)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import os
from copy import deepcopy as dc
import fire
import b3p.blade
import b3p.loft_utils
from ruamel import yaml
import numpy as np


def build_blade_geometry(config, verbose=False, xfoil=False):
    """
    Perform blade 3d model generation based on b3p dictionary.

    :param config: b3p dictionary
    """
    pln = config["planform"]

    blade = b3p.blade.blade(
        pln["chord"],
        pln["thickness"],
        pln["twist"],
        dc(pln["dx"]),
        pln["dy"],
        pln["z"],
        config["aero"]["airfoils"],
        chordwise_sampling=b3p.loft_utils.optspace(config["planform"]["npchord"]),
        np_spanwise=config["planform"]["npspan"],
    )

    wdp = os.path.join(config["general"]["workdir"], config["general"]["prefix"])

    if not os.path.isdir(config["general"]["workdir"]):
        os.makedirs(config["general"]["workdir"])

    blade.mesh(f"{wdp}.vtp")
    blade.dump(f"{wdp}.pck", z_rotation=0)
    blade.export_variables(f"{wdp}.var")
    if xfoil:
        blade.export_xfoil(
            prefix=os.path.join(
                config["general"]["workdir"],
                "airfoil_out",
                f'xs_{config["general"]["prefix"]}',
            )
        )
    blade.plot(f"{wdp}", fname=f"{wdp}.png")
    n_sections = 50
    blade.to_table(np.linspace(0, 1, n_sections), "%s_sca_%i" % (wdp, n_sections))
    return blade


def run_blade_geometry(yamlfile, verbose=False):
    """Build a blade geometry from a yaml file"""
    config = yaml.load(open(yamlfile, "r"), Loader=yaml.CLoader)
    _ = build_blade_geometry(config, verbose=verbose)


def main():
    fire.Fire(run_blade_geometry)



import vtk
from b3p import geometry_section
import numpy
from b3p import geom_utils


class blade_shape:
    def __init__(
        self,
        sections=None,
        section_resolution=200,
        web_resolution=20,
        added_datums=None,
    ):
        if sections is None:
            sections = []
        if added_datums is None:
            added_datums = {}
        self.set_sections(sections)
        self.set_section_resolution(section_resolution)
        self.webs = []
        self.web_resolution = web_resolution
        self.added_datums = added_datums

    def set_web(self, web):
        self.webs.append(web)

    def set_sections(self, sections):
        self.sections = sections

    def set_section_resolution(self, n_points):
        "set the number of points used to sample the airfoil section"
        self.n_points = n_points

    def build_interpolated_sections(self, radii, interpolation_type=1):
        """
        Create interpolated sections

        args:
            interpolation_type (int) : 1==cardinal spline, 2==linear, 3=kochanekspline
        """
        r = [i.r for i in self.sections]
        # get the resplined sections
        pnts = [i.respline(self.n_points)[0] for i in self.sections]

        # define n_points splines along the blade
        nxyz = []
        for i in zip(*pnts):
            if interpolation_type == 1:
                nx = geom_utils.spline_interp(r, list(zip(*i))[0], radii)
                ny = geom_utils.spline_interp(r, list(zip(*i))[1], radii)
                nz = geom_utils.spline_interp(r, list(zip(*i))[2], radii)
            elif interpolation_type == 2:
                nx = numpy.interp(radii, r, list(zip(*i))[0])
                ny = numpy.interp(radii, r, list(zip(*i))[1])
                nz = numpy.interp(radii, r, list(zip(*i))[2])
            elif interpolation_type == 3:
                nx = geom_utils.spline_interp_k(r, list(zip(*i))[0], radii)
                ny = geom_utils.spline_interp_k(r, list(zip(*i))[1], radii)
                nz = geom_utils.spline_interp_k(r, list(zip(*i))[2], radii)

            nxyz.append(list(zip(nx, ny, nz)))

        self.interp_sections = [
            geometry_section.section(i[0], i[0] / max(radii), i[1])
            for i in zip(radii, zip(*nxyz))
        ]

    def mesh(self, n_points=100, close=True, panel_mesh_scale=None):
        """
        Mesh generation for blade

        args:
            n_points (int) : number of points around cross section

            close (bool) : close two ends of mesh (i.e. TE)
        """
        if panel_mesh_scale is None:
            panel_mesh_scale = []
        self.poly = vtk.vtkPolyData()
        points = vtk.vtkPoints()

        # this dict is used to create a number of vtkFloatArrays that have named
        # variables as defined on the section, these can be used to define ply
        # positions
        added_arrays = {}
        # loop over the interpolated sections
        for i in self.interp_sections:
            # respline the section with the standard number of points
            pnts, data = i.respline(
                n_points, self.webs, self.added_datums, panel_mesh_scale
            )
            for j in pnts:
                points.InsertNextPoint(j[0], j[1], j[2])

            for j in data:
                if j not in added_arrays:
                    added_arrays[j] = vtk.vtkFloatArray()
                    added_arrays[j].SetName(j)
                for k in data[j]:
                    added_arrays[j].InsertNextValue(k)

        self.poly.SetPoints(points)

        np = self.poly.GetNumberOfPoints()

        quads = vtk.vtkCellArray()
        for i in range(1, int(np / n_points)):
            np = range((i - 1) * n_points, i * n_points)  # previous row point ids
            nc = range(i * n_points, (i + 1) * n_points)  # current row point ids
            for j in range(n_points - (0 if close else 1)):
                quads.InsertNextCell(4)
                quads.InsertCellPoint(np[j])
                quads.InsertCellPoint(nc[j])
                quads.InsertCellPoint(nc[(j + 1) % n_points])
                quads.InsertCellPoint(np[(j + 1) % n_points])

        self.poly.SetPolys(quads)
        for value in added_arrays.values():
            self.poly.GetPointData().AddArray(value)
        for i in self.webs:
            i.mesh(self.poly, self.web_resolution)

    def write_mesh(self, filename):
        """
        Export mesh in vtp format

        args:
            filename (str): vtp output filename
        """
        try:
            writer = vtk.vtkXMLPolyDataWriter()
            writer.SetFileName(filename)
            writer.SetInputData(self.poly)
            writer.Write()
        except Exception:
            print("no valid mesh available")

        for i in self.webs:
            i.write_mesh(f'{i.name.replace(".txt", ".vtp")}')



import vtk
import math


def spline_interp(x, y, newx):
    spl = vtk.vtkCardinalSpline()
    spl.SetLeftConstraint(2)
    spl.SetRightConstraint(2)
    for i in zip(x, y):
        spl.AddPoint(i[0], i[1])

    return [spl.Evaluate(i) for i in newx]


def spline_interp_k(x, y, newx):
    spl = vtk.vtkKochanekSpline()
    spl.SetLeftConstraint(2)
    spl.SetRightConstraint(2)
    spl.SetDefaultTension(0.0)
    spl.SetDefaultContinuity(0.2)
    for i in zip(x, y):
        spl.AddPoint(i[0], i[1])

    return [spl.Evaluate(i) for i in newx]


def distance(point1, point2):
    return math.sqrt(sum((i[1] - i[0]) ** 2 for i in zip(point1, point2)))



#! /usr/bin/env python3
import pyvista as pv
import sys
import numpy as np
import fire

if pv.OFF_SCREEN:
    pv.start_xvfb()


def visualize_slab_thickness(vtu_file, output_image=None):
    """Create a plot with the slab thicknesses from a VTU file displayed on a mesh.

    Args:
        vtu_file (_type_): grid with laminates
        output_image (_type_, optional): _description_. Defaults to None.
    """
    if output_image is None:
        output_image = vtu_file.replace(".vtu", "_slabs.png")
    # Read the unstructured grid from the VTU file
    grid = pv.read(vtu_file)

    egrid = grid.extract_geometry()

    # Extract the slab arrays
    slab_arrays = [name for name in grid.cell_data.keys() if name.startswith("slab_")]

    if not slab_arrays:
        print("No arrays with names starting with 'slab_' were found in the VTU file.")
        sys.exit(1)

    # Create a plotter object
    plotter = pv.Plotter(off_screen=True)

    # Define the x-offset for the horizontal stacking
    x_offset = 1.3 * abs(grid.bounds[2] - grid.bounds[3])
    # 1.5 times the z-size of the original grid

    # Loop over the slab arrays and add them to the plotter with the appropriate offset and rotation
    for i, slab_name in enumerate(slab_arrays):
        slab_grid = egrid.copy(deep=True)
        slab_grid.cell_data[slab_name] = grid.cell_data[slab_name]
        # Set the active scalars to the current slab array
        slab_grid.rotate_y(
            90, inplace=True
        )  # Rotate the grid around the y-axis by 90 degrees

        if "web" in slab_name:
            non_zero_mask = slab_grid.cell_data[slab_name] != 0
            slab_grid = slab_grid.extract_cells(non_zero_mask)
            slab_grid.rotate_x(
                90, inplace=True
            )  # Rotate the grid around the y-axis by 90 degrees

        slab_grid.translate(
            [0, i * x_offset, 0], inplace=True
        )  # Apply horizontal stacking

        plotter.add_mesh(
            slab_grid,
            show_edges=False,
            cmap="viridis",
            clim=None,
            scalars=slab_name,
            scalar_bar_args={
                "title": slab_name.split("thickness_")[-1],
                "position_x": 0.85,
                "position_y": 0.05 + (i * 0.9 / len(slab_arrays)),
                "width": 0.1,
                "height": 0.1,
                "color": "black",
                "n_labels": 2,
                "title_font_size": 22,
            },
        )

    # Set the camera position
    plotter.view_xy()
    plotter.camera.zoom(1.3)
    # Save the high-resolution image
    plotter.screenshot(
        output_image, transparent_background=True, window_size=[1920, 1080]
    )
    print(f"Saved the high-resolution image to: {output_image}")


def main():
    fire.Fire(visualize_slab_thickness)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import os
from copy import deepcopy as dc
from b3p import mesh_from_loft
from b3p import webs
from ruamel import yaml
import fire
import numpy as np


def build_blade_structure(config):
    """build a rectangle blade mesh with webs, and save it to a vtp file

    :param config: b3p dictionary
    :type config: dict"""
    wdp = os.path.join(config["general"]["workdir"], config["general"]["prefix"])
    pckfile = f"{wdp}.pck"
    base_vtp = f"{wdp}_base.vtp"
    shell_vtp = f"{wdp}_shell.vtp"

    radii = eval(str(config["mesh"]["radii"]))
    config_webs = config["mesh"]["webs"]
    panel_mesh_scale = (
        config["mesh"]["panel_mesh_scale"]
        if "panel_mesh_scale" in config["mesh"]
        else []
    )

    # build the plain blade mesh
    mesh_from_loft.build_mesh(pckfile, radii, [], [], wdp, outfile=base_vtp)

    # compute the locations where the webs intersect with the blade mesh
    web_shell_intersections = webs.build_webs(base_vtp, config_webs, prefix=wdp)

    # rejoggle the datums format from the yaml file into dict
    if "coordinates" in config["mesh"]:
        ad = config["mesh"]["coordinates"]
        added_datums = dict(
            [(i, [ad[i]["base"]] + list(zip(*ad[i]["points"]))) for i in ad]
        )
    else:
        added_datums = {}

    # build the blade mesh with webs
    return mesh_from_loft.build_mesh(
        pckfile,
        radii,
        config_webs,
        web_shell_intersections,
        wdp,
        config["mesh"]["n_web_points"],
        config["mesh"]["n_chordwise_points"],
        outfile=shell_vtp,
        added_datums=added_datums,
        panel_mesh_scale=panel_mesh_scale,
    )


def run_struct(yamlfile):
    """run the build_blade_structure function from the command line
    :param yamlfile: path to the yaml file"""
    bladeconfig = yaml.round_trip_load(open(yamlfile, "r"))
    build_blade_structure(bladeconfig)


def main():
    fire.Fire(run_struct)



#! /usr/bin/env python

import pyvista as pv
from matplotlib import pyplot as plt
import numpy as np
import fire
import os
import json


def plot_thresholded(sm, ax, title, material_map=None):
    p = sm.cell_centers().points
    # get ply arrays
    pk = [i for i in sm.cell_data.keys() if i.startswith("ply_")]
    # get material ids
    mat = [int(sm.cell_data[i][:, 0].max()) for i in pk]

    pt = np.stack([sm.cell_data[i][:, 1] for i in pk])

    base = np.zeros(pt.shape[1])
    for i in range(pt.shape[0]):
        if mat[i] in material_map:
            ax.fill_between(
                p[:, 2],
                y1=base,
                y2=base + pt[i, :],
                label=f"{material_map[mat[i]]}",
                color=f"C{max(mat[i],0)}",
                alpha=0.3,
            )
            base += pt[i, :]

    ax.set_title(title)
    # get unique labels into legend
    handles, labels = ax.get_legend_handles_labels()
    temp = dict(zip(labels, handles))
    ax.legend(temp.values(), temp.keys(), loc="best")


def drape_plot(meshname, output=None):
    """plot the cross section of the laminate plan on (lengthwise) slices of the blade

    :param mesh: path to the mesh in vtu format, needs to have ply_ arrays in cell data
    """
    assert meshname.endswith(".vtu")
    # build the blade mesh
    mesh = pv.read(meshname)

    if output is None:
        output = meshname.replace(".vtu", "_drapeplot.png")

    mm = json.load(open(os.path.join(os.path.dirname(meshname), "material_map.json")))

    imm = {v: k for k, v in mm.items()}
    imm[-1] = "adhesive"
    # if os.path.exists(os.path.join(os.path.dirname(mesh), 'material_map.json')) is False:
    #     os.remove(output)
    # print(mm)

    fig, ax = plt.subplots(4, 1, figsize=(30, 20))

    cnt = mesh.contour(
        scalars="d_w1",
        isosurfaces=[0.0],
    )

    plot_thresholded(cnt, ax[0], title="sparcap", material_map=imm)

    cnt = mesh.contour(
        scalars="d_te",
        isosurfaces=[0.3],
    )

    plot_thresholded(cnt, ax[1], title="trailing edge ud", material_map=imm)

    cnt = mesh.contour(
        scalars="d_w2_r",
        isosurfaces=[0.01],
    )

    plot_thresholded(cnt, ax[2], title="leading edge panel", material_map=imm)

    cnt = mesh.contour(
        scalars="d_w4_r",
        isosurfaces=[-0.05],
    )

    plot_thresholded(cnt, ax[3], title="trailing edge panel", material_map=imm)

    fig.tight_layout()

    fig.savefig(output)
    print(f"written output to {output}")


def main():
    fire.Fire(drape_plot)


if __name__ == "__main__":
    main()



import vtk
import numpy as np
from b3p import geom_utils
import pyvista as pv
import math


def equals(v1, v2):
    tol = 1e-6
    return (v1 - v2) ** 2 < tol


def mesh_line(pnt1, pnt2, n_cells, id):
    """
    utility to mesh a line, adds a couple of parametric coordinates to aid
    draping
    """
    xyz = []
    web_height = vtk.vtkGeoMath().DistanceSquared(pnt1, pnt2) ** 0.5

    for i in zip(pnt1, pnt2):
        mm = min(0.3, 0.06 / web_height)
        rel = sorted([0, 1] + list(np.linspace(mm, 1.0 - mm, n_cells - 2)))
        ab = [j * (i[1] - i[0]) + i[0] for j in rel]
        xyz.append(np.array(ab))

    dst = [i[1:] - i[:-1] for i in xyz]  # distances between points in 3 dimensions
    sl = (dst[0] ** 2 + dst[1] ** 2 + dst[2] ** 2) ** 0.5  # length of the line segments

    # path location from the first web point
    pl = [0] + [sum(sl[:i]) for i in range(1, len(sl) + 1)]

    ppl = [-i + pl[-1] for i in pl]
    ml = [abs(i - 0.5 * web_height) for i in pl]  # distance from the web centerline

    wh = [web_height for _ in ml]

    rad = np.mean(xyz[2])
    r = [rad for _ in range(n_cells)]

    arrays = {
        "d_te": pl,
        "d_le": ppl,
        "d_le_r": [i / max(ppl) for i in ppl],
        f"d_{id}_r": [i / max(ml) for i in ml],
        f"d_{id}": ml,
        # "d_along_airfoil": ml,
        "web_height": wh,
        "radius": r,
        "is_web": [1.0 for _ in ppl],
    }

    return list(zip(*xyz)), arrays


class web:
    def __init__(
        self, points, web_root, web_tip, web_name, coordinate, flip_normal=False
    ):
        self.points = points
        self.web_root = web_root
        self.web_tip = web_tip
        spl1, spl2 = vtk.vtkSCurveSpline(), vtk.vtkSCurveSpline()
        for i in points:
            spl1.AddPoint(i[0], i[1])
            spl2.AddPoint(i[0], i[2])

        self.splines = (spl1, spl2)
        self.evaluations = {}
        self.name = web_name
        self.coordinate = coordinate
        # self.normal = normal
        self.flip_normal = flip_normal

    def average_splits(self):
        """
        routine to define the average of the split position (averaged over
        radius), this is used to calculate the number of points for a shell
        part (which can't be done on the local split positions, since then it
        would vary over R and require the ability to drop and gain element
        strips)
        """
        g = list(zip(*self.points))
        return np.mean(g[1]), np.mean(g[2])

    def splits(self, r, r_relative):
        """
        a split is a point at which the airfoil section has a set point where
        there needs to be a spline evaluation, this ensures that there is a line
        of nodes on the shell to which the web can be attached (or at least
        lined up)
        """
        out = (0, 0)
        out = (self.splines[0].Evaluate(r), self.splines[1].Evaluate(r))
        # log the evaluations of the web position, so that it can be used later
        # to look up the 3D coordinates, store in mm, so that it can be used as
        # an integer key to look up corresponding web split locations
        self.evaluations[int(round(r * 1e2) * 10)] = [out]
        return out

    def _find_top_and_bottom_points(self, mesh):
        """
        loop through the mesh (which represents a shell), when it has been
        constructed to accomodate this web, it will have points on the shell
        where the web starts and ends, this routine finds those points for the
        radius locations where the web is. Note that the length of the web is
        only exact down to the element size

        parameters:
        -----------
        mesh: vtkUnstructuredGrid
            shell mesh to find web points in
        """
        rad = mesh.GetPointData().GetArray("radius")
        rel_dist = mesh.GetPointData().GetArray("d_rel_dist_from_te")
        for i in range(mesh.GetNumberOfPoints()):
            rm = rad.GetValue(i)
            if self.web_root <= rm <= self.web_tip:
                rd = rel_dist.GetValue(i)
                pnt = mesh.GetPoint(i)
                rmm = int(round(rm * 1e2) * 10)
                if equals(rd, self.evaluations[rmm][0][0]) or equals(
                    rd, self.evaluations[rmm][0][1]
                ):
                    self.evaluations[rmm].append(pnt)

    def _create_quad_connectivity(self, n_points, n_total, flip=False):
        """create quad element connectivity array

        parameters:
        -----------
        n_points: int
            number of points along web
        n_total: int
            total number of points in mesh
        returns:
            connectivity array"""
        nrows = int(n_total / n_points)
        colids = np.arange(n_points - 1)
        npp = (
            np.arange(1, nrows).repeat(n_points - 1).reshape(nrows - 1, n_points - 1)
            - 1
        ) * n_points + colids
        if flip:
            stck = [
                np.ones_like(npp) * 4,
                npp + 1,
                npp + n_points + 1,
                npp + n_points,
                npp,
            ]
        else:
            stck = [
                np.ones_like(npp) * 4,
                npp,
                npp + n_points,
                npp + n_points + 1,
                npp + 1,
            ]
        return np.stack(stck).T.flatten()

    def _create_points(self, n_cells):
        """
        generate the points needed to build the mesh
        """
        ev = self.evaluations
        vp = []
        added_arrays = {}
        for i in sorted(ev):
            if len(ev[i]) == 3:
                pnts, data = mesh_line(ev[i][1], ev[i][2], n_cells, self.coordinate)
                vp.extend(pnts)
                for j in data:
                    if j not in added_arrays:
                        added_arrays[j] = data[j]
                    else:
                        added_arrays[j].extend(data[j])

        return vp, added_arrays

    def write_mesh(self, vtpfile):
        self.mesh.save(vtpfile)
        print(f"** wrote mesh to {vtpfile}")

    def mesh(self, mesh, n_cells):
        """
        create the mesh for the web

        parameters:
        -----------
        mesh: vtkUnstructuredGrid
            shell mesh to find web points in
        n_cells: int
            number of cells to use to represent the web
        """
        self._find_top_and_bottom_points(mesh)
        points, pdata = self._create_points(n_cells)

        cells = self._create_quad_connectivity(n_cells, len(points), self.flip_normal)

        # print(cells)
        # ccc = cells[::5]

        self.mesh = pv.PolyData(points, faces=cells)

        for i in pdata:
            self.mesh.point_data[i] = np.array(pdata[i]).astype(np.float32)



#! /usr/bin/env python3
import pyvista as pv
import argparse
import multiprocessing


def conv3d_2d(mesh):
    """hack to translate a mesh to 2D format, which is what fenics/anba needs
    :param mesh: path to the mesh file
    """
    x = open(mesh, "r").read()
    nb = x[: x.find("Topology")]
    rest = x[x.find("Topology") :]
    nb = nb.replace("0.0000000e+00\n", "\n")
    nb = nb.replace(' 3"', ' 2"')
    nb = nb.replace("XYZ", "XY")
    open(mesh, "w").write(nb + rest)


def vtp2xdmf(vtp):
    """convert a vtp file to xdmf format and translate to 2D"""
    assert vtp.endswith(".vtp")
    mesh = pv.read(vtp)
    tri = mesh.triangulate()  # fenics doesn't do mixed element types
    tri.points[:, 2] = 0
    xd = vtp.replace(".vtp", ".xdmf")
    pv.save_meshio(xd, tri, data_format="XML")
    conv3d_2d(xd)
    print(f"converted {vtp} to {xd}")


def anba4_prep(section_meshes):
    p = multiprocessing.Pool()
    p.map(vtp2xdmf, section_meshes)
    p.close()


def main():
    p = argparse.ArgumentParser(
        description="translate section meshes from vtk to XDMF and 2D"
    )
    p.add_argument("sections", nargs="*", help="section meshes in .vtp format")
    args = p.parse_args()
    anba4_prep(args.sections)


if __name__ == "__main__":
    main()



import vtk
import numpy as np
import math
import os
import pyvista as pv


class section:
    def __init__(self, x, y):
        self.x, self.y = x, y
        pnts = np.column_stack((x, y, np.zeros_like(x)))
        cells = np.column_stack(
            (
                2 * np.ones(len(x)).astype(int),
                np.arange(0, len(x)),
                np.arange(1, (len(x) + 1)) % len(x),
            )
        ).flatten()
        self.polydata = pv.PolyData(pnts, lines=cells)

    def local_to_global(self):
        """Transform the section to global coordinates from airfoil coordinates"""
        self.polydata.points = np.array(
            [self.polydata.points[:, i] for i in [1, 0, 2]]
        ).T

    def get_max_thickness(self, web_angle=0, n_points=50):
        """
        get the location (x) where the airfoil is thickest, used to offset the
        section for maximum building height.
        """
        bounds = self.polydata.GetBounds()
        dx = bounds[1] - bounds[0]
        px = np.linspace(bounds[0] + 0.15 * dx, bounds[1] - 0.4 * dx, n_points)
        plane = vtk.vtkPlane()
        clip = vtk.vtkCutter()

        clip.SetInputData(self.polydata)
        plane.SetNormal(
            np.cos(math.radians(web_angle)), math.sin(math.radians(web_angle)), 0
        )
        t = []
        tb = []
        for i in px:
            plane.SetOrigin(i, 0.5 * (bounds[2] + bounds[3]), 0)
            clip.SetCutFunction(plane)
            clip.Update()
            section = clip.GetOutput()
            top, bot = section.GetPoint(0), section.GetPoint(1)
            t.append(section.GetPoint(0)[1] - section.GetPoint(1)[1])
            tb.append((top, bot))

        return px[t.index(max(t))]

    def scale(self, scalefactor):
        self.polydata.scale(scalefactor, inplace=True)

    def twist(self, rz):
        self.polydata.rotate_z(rz, inplace=True)

    def translate(self, dx, dy, dz):
        self.polydata.translate([dx, dy, dz], inplace=True)

    def get_point(self, xy):
        return self.polydata.GetPoint(self.polydata.FindPoint((xy[0], xy[1], 0.0)))

    def get_pointlist(self, z_rotation=0):
        output = self.polydata.rotate_z(z_rotation, inplace=False)
        return output.points

    def to_xfoil(self, fname):
        if not os.path.isdir(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        with open(fname, "w") as f:
            for i in zip(self.x, self.y):
                f.write("%f      %f\n" % (i[0], i[1]))

    def get_te(self):
        te1 = self.polydata.GetPoint(0)
        te2 = self.polydata.GetPoint(self.polydata.GetNumberOfPoints() - 1)
        return (
            [0.5 * (i[0] + i[1]) for i in zip(te1, te2)],
            vtk.vtkMath.Distance2BetweenPoints(te1, te2) ** 0.5,
        )



#! /usr/bin/env python
import glob
import fire
import os
import pyvista as pv
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from cycler import cycler


def get_quadrant(vector):
    ang = np.degrees(np.arctan2(vector[1], vector[0]))

    if np.isclose(ang, -90, atol=15):
        return -180
    elif np.isclose(ang, 90, atol=15):
        return 0
    elif np.isclose(ang, 0, atol=15):
        return 90
    elif np.isclose(ang, 180, atol=15):
        return -90
    else:
        return 45


class plot_ccx:
    def __init__(self, wdir, wildcard=""):
        self.wdir = wdir
        self.meshes = glob.glob(wdir + f"/*ccx*{wildcard}*.vtu")

    def plot3d(self):
        for i in self.meshes:
            mesh = pv.read(i)
            self.__plot3d(mesh, output_path=i.replace(".vtu", "_3d.png"))
            del mesh
        return self

    def __plot3d(self, mesh, output_path):
        ts = np.unique([i.split("_")[1] for i in mesh.point_data.keys()])

        for i in ts:
            isdisp = i.startswith("0.00000")
            # Deform the mesh using the ChannelDisplacement filter
            deform = mesh.point_data[f"disp_{ts[0]}"]
            deformed_mesh = mesh.warp_by_vector(
                f"disp_{ts[0]}", factor=1  # if isdisp else None
            )

            if isdisp:  # f"strain_{i}" in mesh.point_data.keys():
                # Color the mesh using the third component of point data strain
                deformed_mesh.point_data["strain"] = mesh.point_data[f"strain_{i}"][
                    :, 2
                ]
                deformed_mesh.set_active_scalars("strain")
            else:
                # Color the mesh using the third component of point data strain
                deformed_mesh.point_data["disp"] = mesh.point_data[f"disp_{i}"].sum(
                    axis=1
                )
                deformed_mesh.set_active_scalars("disp")

            nrm = np.linalg.norm(deform, axis=1)

            whr = np.where(nrm == nrm.max())
            quadrant = get_quadrant(deform[whr[0][0]][:2])

            # print(quadrant, output_path, deform[whr[0][0]][:2])
            print(quadrant)
            deformed_mesh.rotate_z(quadrant, inplace=True)

            deformed_mesh.rotate_y(-110, inplace=True)

            # Create a plot in isometric view
            plotter = pv.Plotter(off_screen=True, window_size=[1920, 600])
            plotter.add_mesh(
                deformed_mesh,
                cmap="coolwarm",
                clim=[-5e-3, 5e-3] if isdisp else None,
                scalar_bar_args={
                    "position_x": 0.1,
                    "position_y": 0.05,
                    "width": 0.8,
                    "height": 0.06,
                    "color": "black",
                    "n_labels": 2,
                    "title_font_size": 22,
                },
            )
            plotter.view_isometric()
            plotter.camera.parallel_projection = True
            plotter.camera.zoom(3.5)

            # Save the plot to a high-resolution image
            of = output_path.replace(".png", f"_{i}.png")
            plotter.screenshot(
                of,
                transparent_background=True,
                return_img=False,
            )
            print(f"Saved {of}")
            # clear up memory
            plotter.close()
            del plotter

    def __str__(self):
        return ""

    def plot2d(self, wildcard=""):
        pqs = glob.glob(os.path.join(self.wdir, f"*{wildcard}*eps2d.pq"))

        for pq in pqs:
            df = pd.read_parquet(pq)
            fig, ax = plt.subplots(figsize=(12, 8))
            ax.set_prop_cycle(
                cycler(color=["red", "blue", "green", "orange"])
                + cycler(linestyle=["-", "--", ":", "-."])
            )
            for i in df:
                if i[2] == "zz":
                    ax.plot(
                        df[i].index.values,
                        df[i].values,
                        label=" ".join(i[:2]),
                    )
            ax.set_xlabel("z-coordinate")
            ax.set_ylabel("strain")
            ax.legend()
            ax.set_ylim(-7e-3, 7e-3)
            ax.grid(True)
            output_path = pq.replace(".pq", ".png")
            fig.savefig(output_path, dpi=300)  # , transparent=True)
            print(f"Saved {output_path}")
            plt.close(fig)


if __name__ == "__main__":
    fire.Fire(plot_ccx)  # main()
